import unittest
import os
import tempfile
from math import sin, cos
from propertime.utils import dt

from ..datastructures import TimePoint, DataTimeSlot, DataTimePoint, TimeSeries
from ..models.forecasters import ProphetForecaster, PeriodicAverageForecaster, ARIMAForecaster, AARIMAForecaster, LSTMForecaster
from ..storages import CSVFileStorage
from ..utils import ensure_reproducibility, PFloat

# Setup logging
from .. import logger
logger.setup()

# Test data and temp path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'
TEMP_MODELS_DIR = tempfile.TemporaryDirectory().name


class TestForecasters(unittest.TestCase):

    def setUp(self):

        # Create a minute-resolution test DataTimeSlotSeries
        self.sine_minute_timeseries = TimeSeries()
        for i in range(1000):
            self.sine_minute_timeseries.append(DataTimeSlot(start=TimePoint(i*60), end=TimePoint((i+1)*60), data={'value':sin(i/10.0)}))

        # Create a day-resolution test DataTimeSlotSeries
        self.sine_day_timeseries = TimeSeries()
        for i in range(1000):
            step = 60 * 60 * 24
            self.sine_day_timeseries.append(DataTimeSlot(start=TimePoint(i*step), end=TimePoint((i+1)*step), data={'value':sin(i/10.0)}))

        # Ensure reproducibility
        ensure_reproducibility()


    @staticmethod
    def assertCompatible(actual, predicted, delta=0.3, item=None):
        if not abs(actual-predicted) < delta:
            raise AssertionError('Actual and predicted values are not compatible within the specified delta (actual="{}", predicted="{}", delta="{}", item="{}"'.format(actual, predicted, delta, item))


    def test_PeriodicAverageForecaster(self):

        # Note: here we test also some basic functionality of the forecaster as well,
        # given that the PeridicAVergaeForecaster is very simple and close to a mock.

        # Instantiate and fit
        forecaster = PeriodicAverageForecaster()
        forecaster.fit(self.sine_minute_timeseries, periodicity=63)

        # Apply
        sine_minute_timeseries_with_forecast = forecaster.apply(self.sine_minute_timeseries, steps=3)
        self.assertEqual(len(sine_minute_timeseries_with_forecast), 1003)

        # Predict
        prediction = forecaster.predict(self.sine_minute_timeseries, steps=3)
        self.assertTrue(isinstance(prediction, list))
        self.assertEqual(len(prediction), 3)

        # Evaluate
        evaluation = forecaster.evaluate(self.sine_minute_timeseries[0:100], steps=1)
        self.assertEqual(forecaster.data['periodicities']['value'], 63)
        self.assertAlmostEqual(evaluation['value_RMSE'], 0.0873, places=2)
        self.assertAlmostEqual(evaluation['value_MAE'], 0.0797, places=2)

        # Test on a realistic point series, forecast horizon=1
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = PeriodicAverageForecaster()
        with self.assertRaises(ValueError):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries)

        prediction = forecaster.predict(timeseries)
        self.assertEqual(len(prediction), 1)
        self.assertEqual(list(prediction[0].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[0]['temperature'], 22.891000000000005)

        forecast = forecaster.forecast(timeseries)
        self.assertEqual(len(forecast), 1)
        self.assertEqual(forecast[0].t, timeseries[-1].t+600)
        self.assertEqual(list(forecast[0].data.keys()), ['temperature'])
        self.assertAlmostEqual(forecast[0].data['temperature'], 22.891000000000005)

        timeseries_with_forecast  = forecaster.apply(timeseries)
        self.assertEqual(len(timeseries_with_forecast), len(timeseries)+1)
        self.assertEqual(timeseries_with_forecast[-1].t, timeseries[-1].t+600)
        self.assertEqual(list(timeseries_with_forecast[-1].data.keys()), ['temperature'])
        self.assertAlmostEqual(timeseries_with_forecast[-1].data['temperature'], 22.891000000000005)

        # Test on a realistic point series, forecast horizon=3
        prediction = forecaster.predict(timeseries, steps=3)
        self.assertEqual(len(prediction), 3)
        self.assertEqual(list(prediction[0].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[0]['temperature'], 22.891000000000005)
        self.assertEqual(list(prediction[-1].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[-1]['temperature'], 22.922057142857145)

        forecast = forecaster.forecast(timeseries, steps=3)
        self.assertEqual(len(forecast), 3)
        self.assertEqual(forecast[0].t, timeseries[-1].t+600)
        self.assertEqual(list(forecast[0].data.keys()), ['temperature'])
        self.assertAlmostEqual(forecast[0].data['temperature'], 22.891000000000005)
        self.assertEqual(forecast[-1].t, timeseries[-1].t+600*3)
        self.assertEqual(list(forecast[-1].data.keys()), ['temperature'])
        self.assertAlmostEqual(forecast[-1].data['temperature'], 22.922057142857145)

        timeseries_with_forecast  = forecaster.apply(timeseries, steps=3)
        self.assertEqual(len(timeseries_with_forecast), len(timeseries)+3)
        self.assertEqual(timeseries_with_forecast[-3].t, timeseries[-1].t+600)
        self.assertEqual(list(timeseries_with_forecast[-3].data.keys()), ['temperature'])
        self.assertAlmostEqual(timeseries_with_forecast[-3].data['temperature'], 22.891)
        self.assertEqual(timeseries_with_forecast[-1].t, timeseries[-1].t+600*3)
        self.assertEqual(list(timeseries_with_forecast[-1].data.keys()), ['temperature'])
        self.assertAlmostEqual(timeseries_with_forecast[-1].data['temperature'], 22.922057142857145)


    def test_PeriodicAverageForecaster_multivariate(self):

        # Note: here we test also some basic functionality of the forecaster as well,
        # given that the PeridicAVergaeForecaster is very simple and close to a mock.

        timeseries = TimeSeries()
        for i in range(970):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0), 'cos': cos(i/10.0)}))

        # Instantiate and fit
        forecaster = PeriodicAverageForecaster()
        forecaster.fit(timeseries)

        prediction = forecaster.predict(timeseries, steps=3)
        self.assertEqual(len(prediction), 3)
        self.assertEqual(list(prediction[0].keys()), ['cos', 'sin'])
        self.assertAlmostEqual(prediction[0]['cos'], 0.19600497357083713)
        self.assertAlmostEqual(prediction[0]['sin'], 0.660933759485128)
        self.assertEqual(list(prediction[-1].keys()), ['cos', 'sin'])
        self.assertAlmostEqual(prediction[-1]['cos'], 0.0642538467366801)
        self.assertAlmostEqual(prediction[-1]['sin'], 0.6923761210530861)

        forecast = forecaster.forecast(timeseries, steps=3)
        self.assertEqual(len(forecast), 3)
        self.assertEqual(forecast[0].t, timeseries[-1].t+60)
        self.assertEqual(list(forecast[0].data.keys()), ['cos', 'sin'])
        self.assertAlmostEqual(forecast[0].data['cos'], 0.19600497357083713)
        self.assertAlmostEqual(forecast[0].data['sin'], 0.660933759485128)
        self.assertEqual(forecast[-1].t, timeseries[-1].t+60*3)
        self.assertEqual(list(forecast[-1].data.keys()), ['cos', 'sin'])
        self.assertAlmostEqual(forecast[-1].data['cos'], 0.0642538467366801)
        self.assertAlmostEqual(forecast[-1].data['sin'], 0.6923761210530861)

        timeseries_with_forecast  = forecaster.apply(timeseries, steps=3)
        self.assertEqual(len(timeseries_with_forecast), len(timeseries)+3)
        self.assertEqual(timeseries_with_forecast[-3].t, timeseries[-1].t+60)
        self.assertEqual(list(timeseries_with_forecast[-3].data.keys()), ['cos', 'sin'])
        self.assertAlmostEqual(timeseries_with_forecast[-3].data['cos'], 0.19600497357083713)
        self.assertAlmostEqual(timeseries_with_forecast[-3].data['sin'], 0.660933759485128)
        self.assertEqual(timeseries_with_forecast[-1].t, timeseries[-1].t+60*3)
        self.assertEqual(list(timeseries_with_forecast[-1].data.keys()), ['cos', 'sin'])
        self.assertAlmostEqual(timeseries_with_forecast[-1].data['cos'], 0.0642538467366801)
        self.assertAlmostEqual(timeseries_with_forecast[-1].data['sin'], 0.6923761210530861)


    def test_PeriodicAverageForecaster_probabilistic(self):

        # Instantiate and fit
        forecaster = PeriodicAverageForecaster()
        forecaster.fit(self.sine_minute_timeseries, periodicity=63, probabilistic=True)

        # Check internal data
        self.assertIsInstance(forecaster.data['offsets_averages']['value'][0], PFloat)

        # Predict
        predicted = forecaster.predict(self.sine_minute_timeseries, steps=1)
        self.assertIsInstance(predicted[0]['value'], PFloat)


    def test_PeriodicAverageForecaster_save_load(self):

        forecaster = PeriodicAverageForecaster()
        forecaster.fit(self.sine_minute_timeseries, periodicity=63)
        forecaster.save(TEMP_MODELS_DIR + '/test_PA_model')

        loaded_forecaster = PeriodicAverageForecaster.load(TEMP_MODELS_DIR + '/test_PA_model')
        self.assertEqual(forecaster.data['offsets_averages'], loaded_forecaster.data['offsets_averages'])


    def test_ProphetForecaster(self):

        try:
            import prophet
        except ImportError:
            print('Skipping Prophet tests as no prophet module installed')
            return

        forecaster = ProphetForecaster()

        forecaster.fit(self.sine_day_timeseries)
        self.assertEqual(len(self.sine_day_timeseries), 1000)

        sine_day_timeseries_with_forecast = forecaster.apply(self.sine_day_timeseries, steps=3)
        self.assertEqual(len(sine_day_timeseries_with_forecast), 1003)

        # Test with the evaluate
        evalation_results = forecaster.evaluate(self.sine_day_timeseries.view(0,10))
        self.assertAlmostEqual(evalation_results['value_RMSE'], 0.80, places=2)
        self.assertAlmostEqual(evalation_results['value_MAE'], 0.77, places=2)

        evalation_results = forecaster.evaluate(self.sine_day_timeseries.view(0,1))
        self.assertAlmostEqual(evalation_results['value_RMSE'], 0.53, places=2) # For one sample they must be the same
        self.assertAlmostEqual(evalation_results['value_MAE'], 0.53, places=2) # For one sample they must be the same

        # Test on a realistic point series, forecast horizon=3
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = ProphetForecaster()
        with self.assertRaises(Exception):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries)

        prediction = forecaster.predict(timeseries, steps=3)
        self.assertEqual(len(prediction), 3)
        self.assertEqual(list(prediction[0].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[0]['temperature'], 23.297123620317347, places=2)
        self.assertEqual(list(prediction[-1].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[-1]['temperature'], 23.336020073857974, places=2)


    def test_ARIMAForecaster(self):

        try:
            import statsmodels
        except ImportError:
            print('Skipping ARIMA tests as no statsmodels module installed')
            return

        import warnings
        from statsmodels.tools.sm_exceptions import ConvergenceWarning
        warnings.simplefilter('ignore', ConvergenceWarning)

        # Basic ARIMA
        forecaster = ARIMAForecaster(p=1,d=1,q=0)
        forecaster.fit(self.sine_day_timeseries)

        # Cannot apply on a time series ending with a different timestamp as the time series used for the fit
        with self.assertRaises(Exception):
            forecaster.apply(self.sine_day_timeseries[:-1], steps=3)

        sine_day_timeseries_with_forecast = forecaster.apply(self.sine_day_timeseries, steps=3)
        self.assertEqual(len(sine_day_timeseries_with_forecast), 1003)

        # Test on a realistic point series, forecast horizon=3
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = ARIMAForecaster()
        with self.assertRaises(Exception):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries)

        prediction = forecaster.predict(timeseries, steps=3)
        self.assertEqual(len(prediction), 3)
        self.assertEqual(list(prediction[0].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[0]['temperature'], 23.227174207689878)
        self.assertEqual(list(prediction[-1].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[-1]['temperature'], 23.23067079831882)


    def test_AARIMAForecaster(self):

        try:
            import statsmodels
        except ImportError:
            print('Skipping AARIMA tests as no statsmodels module installed')
            return

        # Automatic ARIMA
        forecaster = AARIMAForecaster()

        forecaster.fit(self.sine_day_timeseries, max_p=2, max_d=1, max_q=2)
        self.assertEqual(len(self.sine_day_timeseries), 1000)

        # Cannot apply on a time series ending with a different timestamp as the time series used for the fit
        with self.assertRaises(Exception ):
            forecaster.apply(self.sine_day_timeseries[:-1], steps=3)

        sine_day_timeseries_with_forecast = forecaster.apply(self.sine_day_timeseries, steps=3)
        self.assertEqual(len(sine_day_timeseries_with_forecast), 1003)

        # Test on a realistic point series, forecast horizon=3
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = AARIMAForecaster()
        with self.assertRaises(Exception):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries)

        prediction = forecaster.predict(timeseries, steps=3)
        self.assertEqual(len(prediction), 3)
        self.assertEqual(list(prediction[0].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[0]['temperature'], 23.227174207689878)
        self.assertEqual(list(prediction[-1].keys()), ['temperature'])
        self.assertAlmostEqual(prediction[-1]['temperature'], 23.23067079831882)


    def test_LSTMForecaster_univariate(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with univariate time series as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0)}))

        # Basic test
        forecaster = LSTMForecaster(window=10)
        forecaster.fit(timeseries[0:970], epochs=10, reproducible=True, verbose=False)
        timeseries_forecast = forecaster.apply(timeseries[0:970], steps=10)

        evaluation_results = forecaster.evaluate(timeseries[0:100], error_metrics=['MAE'])
        self.assertTrue(evaluation_results['sin_MAE'] <0.2, evaluation_results['sin_MAE'])

        for i in range(10):
            actual = timeseries[970+i].data['sin']
            predicted = timeseries_forecast[970+i].data['sin']
            self.assertCompatible(actual, predicted)

        # Test using other parameters and features
        forecaster = LSTMForecaster(window=10, neurons=64, features=['values','diffs'])
        forecaster.fit(timeseries[0:970], epochs=10, reproducible=True, verbose=False)
        timeseries_forecast = forecaster.apply(timeseries[0:970], steps=10)

        evaluation_results = forecaster.evaluate(timeseries[0:100], error_metrics=['MAE'])
        self.assertTrue(evaluation_results['sin_MAE'] <0.2, evaluation_results['sin_MAE'])

        for i in range(10):
            actual = timeseries[970+i].data['sin']
            predicted = timeseries_forecast[970+i].data['sin']
            self.assertCompatible(actual, predicted, item=i)

        # Test for non-existent features
        with self.assertRaises(ValueError):
            LSTMForecaster(features=['values','not_existent_feature']).fit(timeseries)


    def test_LSTMForecaster_univariate_with_window_mask(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with univariate time series as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0)}))

        # Non-ngarive mask not supported
        forecaster = LSTMForecaster(window=10, window_mask=3)
        with self.assertRaises(ValueError):
            forecaster.fit(timeseries[0:970], epochs=10, reproducible=True, verbose=False)

        # Negative mask indicates how many elements to dismiss at the end of the window
        forecaster = LSTMForecaster(window=10, window_mask=-3)
        forecaster.fit(timeseries[0:970], epochs=10, reproducible=True, verbose=False)

        evaluation_results = forecaster.evaluate(timeseries[0:100], error_metrics=['MAE'])
        self.assertAlmostEqual(evaluation_results['sin_MAE'], 0.049999925551957244, places=3)


    def test_LSTMForecaster_univariate_probabilistic(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM probabilistic forecaster tests with univariate time series as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0)}))

        forecaster = LSTMForecaster(window=10)
        forecaster.fit(timeseries[0:970], epochs=3, reproducible=True, probabilistic=True, verbose=False)

        predicted = forecaster.predict(timeseries[0:970], samples=3)
        self.assertAlmostEqual(predicted['sin'], 0.593242318431153, places=3)
        self.assertIsInstance(predicted['sin'], PFloat)


    def test_LSTMForecaster_multivariate(self):
        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with multivariate time series as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0), 'cos': cos(i/10.0)}))

        # Basic test
        forecaster = LSTMForecaster()
        forecaster.fit(timeseries, epochs=10, reproducible=True)
        evaluation_results = forecaster.evaluate(timeseries[0:100], error_metrics=['MAE'], return_evaluation_series=True)

        self.assertTrue(evaluation_results['sin_MAE'] < 0.05, evaluation_results['sin_MAE'])
        self.assertTrue(evaluation_results['sin_MAE'] < 0.05, evaluation_results['sin_MAE'])

        timeseries_forecast = forecaster.apply(timeseries[0:970], steps=10)

        for i in range(10):
            actual = timeseries[970+i].data['sin']
            predicted = timeseries_forecast[970+i].data['sin']
            self.assertCompatible(actual, predicted)

        for i in range(10):
            actual = timeseries[970+i].data['cos']
            predicted = timeseries_forecast[970+i].data['cos']
            self.assertCompatible(actual, predicted)


    def test_LSTMForecaster_multivariate_with_targets_and_context(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with multivariate time series with targets as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0), 'cos': cos(i/10.0)}))

        # Target, no context
        forecaster = LSTMForecaster()
        forecaster.fit(timeseries[0:978], epochs=10, target='cos', with_context=False, reproducible=True)

        with self.assertRaises(ValueError):
            forecaster.predict(timeseries[0:978], context_data={'sin': 'whatever'})

        predicted_data = forecaster.predict(timeseries)
        self.assertEqual(list(predicted_data.keys()), ['cos'])
        self.assertCompatible(timeseries[979].data['cos'], predicted_data['cos'])

        # Target and context
        forecaster = LSTMForecaster()
        forecaster.fit(timeseries[0:978], target='cos', with_context=True, reproducible=True)

        with self.assertRaises(ValueError):
            forecaster.predict(timeseries[0:978])

        predicted_data = forecaster.predict(timeseries[0:978], context_data={'sin': timeseries[979].data['cos']})
        self.assertEqual(list(predicted_data.keys()), ['cos'])
        self.assertCompatible(timeseries[979].data['cos'], predicted_data['cos'])

        # All other methods will raise as not suitable
        with self.assertRaises(ValueError):
            forecaster.forecast(timeseries)

        with self.assertRaises(ValueError):
            forecaster.apply(timeseries)


    def test_LSTMForecaster_cross_validation(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster cross validation tests as no tensorflow module installed')
            return

        temperature_timeseries = TimeSeries.from_csv(TEST_DATA_PATH + 'csv/temperature_winter.csv').resample('1h')

        # Pretend there was no data loss at all
        for item in temperature_timeseries:
            item.data_indexes['data_loss'] = 0 

        # Initialize and cross validate the forecaster
        forecaster = LSTMForecaster(window=12, neurons=64, features=['values', 'diffs', 'hours'])
        cross_validation_results = forecaster.cross_validate(temperature_timeseries[0:100], rounds=3, fit_reproducible=True,
                                                             evaluate_return_evaluation_series=True, return_full_evaluations=True)

        self.assertTrue(0.3 < cross_validation_results['temperature_RMSE_avg'] < 0.6) # Expected: 0.45243811980703913
        self.assertTrue(0.2 < cross_validation_results['temperature_MAE_avg'] < 0.5) # Expected: 0.36419170510171595
        self.assertEqual(len(cross_validation_results['evaluations']),3)
        for evaluation in cross_validation_results['evaluations']:
            self.assertTrue('temperature_RMSE' in evaluation)
            self.assertTrue('temperature_MAE' in evaluation)
            self.assertTrue('series' in evaluation)


    def test_LSTMForecaster_save_load(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM save/load forecaster tests as no tensorflow module installed')
            return

        # Create a minute-resolution test DataTimeSlotSeries
        sine_minute_timeseries = TimeSeries()
        for i in range(10):
            sine_minute_timeseries.append(DataTimeSlot(start=TimePoint(i*60), end=TimePoint((i+1)*60), data={'sin':sin(i/10.0), 'cos':cos(i/10.0)}))

        # Fit a forecaster and make a prediction
        forecaster = LSTMForecaster()
        forecaster.fit(sine_minute_timeseries, reproducible=True)
        predicted_data = forecaster.predict(sine_minute_timeseries)

        # Set model path
        model_path = TEMP_MODELS_DIR+ '/test_LSTM_model'

        # Save
        forecaster.save(model_path)

        # Load
        forecaster_loaded = LSTMForecaster.load(model_path)

        # Predict from the loaded model
        predicted_data_loaded = forecaster_loaded.predict(sine_minute_timeseries)

        # Check predictions are the same
        self.assertEqual(predicted_data, predicted_data_loaded)


    def test_LinearRegressionForecaster(self):

        # Create a hour-resolution test DataTimeSlotSeries
        sine_hour_timeseries = TimeSeries()
        for i in range(1000):
            sine_hour_timeseries.append(DataTimePoint(t=i*3600, data={'value':sin(i/10.0)}))

        from ..models import LinearRegressionForecaster
        forecaster = LinearRegressionForecaster(features=['values','hours'])

        # Fit
        forecaster.fit(sine_hour_timeseries)

        # Predict and check
        self.assertAlmostEqual(forecaster.apply(sine_hour_timeseries)[-1].data['value'], -0.5063, places=3)

        # Save
        forecaster.save(TEMP_MODELS_DIR+'/test_lr_model')

        # Load
        loaded_forecaster = LinearRegressionForecaster.load(TEMP_MODELS_DIR+'/test_lr_model')

        # Re-predict and re-check
        self.assertAlmostEqual(loaded_forecaster.apply(sine_hour_timeseries)[-1].data['value'], -0.5063, places=3)




import unittest
import os
import tempfile
from propertime.utils import dt

from ..datastructures import TimePoint, DataTimeSlot, DataTimePoint, TimeSeries
from ..models.reconstructors import PeriodicAverageReconstructor, ProphetReconstructor, LinearInterpolationReconstructor
from ..storages import CSVFileStorage
from ..utils import ensure_reproducibility

# Setup logging
from .. import logger
logger.setup()

# Test data and temp path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'
TEMP_MODELS_DIR = tempfile.TemporaryDirectory().name


class TestReconstructors(unittest.TestCase):

    def setUp(self):

        # Ensure reproducibility
        ensure_reproducibility()


    def test_LinearInterpolationReconstructor(self):

        # Note: here we test also some basic functionality of the reconstructors as well,
        # given that the LinearInterpolationReconstructor is very simple and close to a mock.

        # Test simple univariate
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=1,  data={'a':1}, data_loss=0))
        timeseries.append(DataTimePoint(t=2,  data={'a':-99}, data_loss=1))
        timeseries.append(DataTimePoint(t=3,  data={'a':3}, data_loss=1))
        timeseries.append(DataTimePoint(t=4,  data={'a':4}, data_loss=0))

        reconstructor = LinearInterpolationReconstructor()

        # Test the main predict
        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=1)
        self.assertEqual(predicted_data, [{'a': 2.0}])

        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=2)
        self.assertEqual(predicted_data, [{'a': 2.0}, {'a': 3.0}])

        # Not enough widow data for the predict, will raise
        with self.assertRaises(ValueError):
            reconstructor.predict(timeseries, from_i=0,to_i=1)
        with self.assertRaises(ValueError):
            reconstructor.predict(timeseries, from_i=3,to_i=4)

        # Test the apply now
        reconstructed_timeseries = reconstructor.apply(timeseries)
        self.assertEqual(reconstructed_timeseries[1].data['a'],2)

        # Not enough widow data for the apply at the beginning
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=1,  data={'a':-99}, data_loss=1))
        timeseries.append(DataTimePoint(t=2,  data={'a':2}, data_loss=0))
        with self.assertRaises(ValueError):
            reconstructor.apply(timeseries)

        # Not enough widow data for the apply at the end
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=1,  data={'a':-99}, data_loss=1))
        timeseries.append(DataTimePoint(t=2,  data={'a':2}, data_loss=0))
        with self.assertRaises(ValueError):
            reconstructor.apply(timeseries)

        # Test more complex univariate
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=-3,  data={'value':3}))
        timeseries.append(DataTimePoint(t=-2,  data={'value':2}))
        timeseries.append(DataTimePoint(t=-1,  data={'value':1}))
        timeseries.append(DataTimePoint(t=0,  data={'value':0}))
        timeseries.append(DataTimePoint(t=1,  data={'value':1}))
        timeseries.append(DataTimePoint(t=2,  data={'value':2}))
        timeseries.append(DataTimePoint(t=3,  data={'value':3}))
        timeseries.append(DataTimePoint(t=4,  data={'value':4}))
        timeseries.append(DataTimePoint(t=5,  data={'value':5}))
        timeseries.append(DataTimePoint(t=6,  data={'value':6}))
        timeseries.append(DataTimePoint(t=7,  data={'value':7}))
        timeseries.append(DataTimePoint(t=16, data={'value':16}))
        timeseries.append(DataTimePoint(t=17, data={'value':17}))
        timeseries.append(DataTimePoint(t=18, data={'value':18}))

        # Resample for 3 seconds
        resampled_timeseries = timeseries.resample(3)
        # DataTimePoint @ 0.0 (1970-01-01 00:00:00+00:00) with data "{'value': 0.6666666666666666}" and data_loss="0.0"
        # DataTimePoint @ 3.0 (1970-01-01 00:00:03+00:00) with data "{'value': 3.0}" and data_loss="0.0"
        # DataTimePoint @ 6.0 (1970-01-01 00:00:06+00:00) with data "{'value': 6.0}" and data_loss="0.0"
        # DataTimePoint @ 9.0 (1970-01-01 00:00:09+00:00) with data "{'value': 9.0}" and data_loss="1.0"
        # DataTimePoint @ 12.0 (1970-01-01 00:00:12+00:00) with data "{'value': 12.0}" and data_loss="1.0"
        # DataTimePoint @ 15.0 (1970-01-01 00:00:15+00:00) with data "{'value': 15.0}" and data_loss="0.6666666666666667"

        # Fake data to check the reconstructor is working correctly
        resampled_timeseries[3].data['value']=-9
        resampled_timeseries[4].data['value']=-12

        # Instantiate the reconstructor
        reconstructor = LinearInterpolationReconstructor()

        # Check parametric param just in case
        self.assertFalse(reconstructor._is_parametric())

        # Apply the reconstructor
        reconstructed_timeseries = reconstructor.apply(resampled_timeseries)

        # Check len
        self.assertEqual(len(reconstructed_timeseries), 6)

        # Check for t = 0
        self.assertEqual(reconstructed_timeseries[0].t, 0)
        self.assertAlmostEqual(reconstructed_timeseries[0].data['value'], 0.6666666666)
        self.assertEqual(reconstructed_timeseries[0].data_loss, 0)

        # Check for t = 3
        self.assertEqual(reconstructed_timeseries[1].t, 3)
        self.assertEqual(reconstructed_timeseries[1].data['value'], 3)
        self.assertEqual(reconstructed_timeseries[1].data_loss, 0)

        # Check for t = 6
        self.assertEqual(reconstructed_timeseries[2].t, 6)
        self.assertEqual(reconstructed_timeseries[2].data['value'], 6)
        self.assertEqual(reconstructed_timeseries[2].data_loss, 0)

        # Check for t = 9
        self.assertEqual(reconstructed_timeseries[3].t, 9)
        self.assertEqual(reconstructed_timeseries[3].data['value'], 9)
        self.assertEqual(reconstructed_timeseries[3].data_loss, 1)

        # Check for t = 12
        self.assertEqual(reconstructed_timeseries[4].t, 12)
        self.assertEqual(reconstructed_timeseries[4].data['value'], 12)
        self.assertEqual(reconstructed_timeseries[4].data_loss, 1)

        # Check for t = 15
        self.assertEqual(reconstructed_timeseries[5].t, 15)
        self.assertEqual(reconstructed_timeseries[5].data['value'], 15)
        self.assertAlmostEqual(reconstructed_timeseries[5].data_loss, 0.6666666666)

        # Test multivariate
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=1,  data={'a':1, 'b':4}, data_loss=0))
        timeseries.append(DataTimePoint(t=2,  data={'a':-99, 'b':-99}, data_loss=1))
        timeseries.append(DataTimePoint(t=3,  data={'a':3, 'b':2}, data_loss=1))
        timeseries.append(DataTimePoint(t=4,  data={'a':4, 'b':1}, data_loss=0))

        reconstructor = LinearInterpolationReconstructor()

        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=1)
        self.assertEqual(predicted_data, [{'a': 2.0, 'b': 3.0}])

        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=2)
        self.assertEqual(predicted_data, [{'a': 2.0, 'b': 3.0}, {'a': 3.0, 'b': 2.0}])

        reconstructed_timeseries = reconstructor.apply(timeseries)
        self.assertEqual(reconstructed_timeseries[0].data, {'a': 1, 'b': 4})
        self.assertEqual(reconstructed_timeseries[1].data, {'a': 2.0, 'b': 3.0})
        self.assertEqual(reconstructed_timeseries[2].data, {'a': 3.0, 'b': 2.0})
        self.assertEqual(reconstructed_timeseries[3].data, {'a': 4, 'b': 1})
        self.assertEqual(reconstructed_timeseries[0].data_indexes, {'data_loss': 0, 'data_reconstructed': 0})
        self.assertEqual(reconstructed_timeseries[1].data_indexes, {'data_loss': 1, 'data_reconstructed': 1})
        self.assertEqual(reconstructed_timeseries[2].data_indexes, {'data_loss': 1, 'data_reconstructed': 1})
        self.assertEqual(reconstructed_timeseries[3].data_indexes, {'data_loss': 0, 'data_reconstructed': 0})


    def test_PeriodicAverageReconstructor(self):

        # Get test data
        with open(TEST_DATA_PATH + '/csv/temp_slots_1h.csv') as f:
            data=f.read()

        timeseries = TimeSeries()
        for line in data.split('\n'):
            if line:
                start_t = float(line.split(',')[0])
                start_point = TimePoint(t=start_t)
                end_point = TimePoint(t=start_t+3600)
                data_loss = 1-float(line.split(',')[2]) # from coverage to data loss
                value =  float(line.split(',')[1])
                timeseries.append(DataTimeSlot(start=start_point, end=end_point, data={'temperature':value}, data_loss=data_loss))

        # Fit
        reconstructor = PeriodicAverageReconstructor()
        reconstructor.fit(timeseries)

        # Test the evaluate
        evaluation = reconstructor.evaluate(timeseries, limit=100, details=True)
        self.assertAlmostEqual(evaluation['RMSE_1_steps'], 0.131, places=2)
        self.assertAlmostEqual(evaluation['MAE_1_steps'], 0.094, places=2)
        self.assertAlmostEqual(evaluation['RMSE_24_steps'], 0.641, places=2)
        self.assertAlmostEqual(evaluation['MAE_24_steps'], 0.526, places=2)
        self.assertAlmostEqual(evaluation['RMSE'], 0.386, places=2)
        self.assertAlmostEqual(evaluation['MAE'], 0.310, places=2)

        # Test the evaluate on specific steps
        evaluation = reconstructor.evaluate(timeseries, steps=[1,3], limit=100, details=True)
        self.assertAlmostEqual(evaluation['RMSE_1_steps'], 0.131, places=2)
        self.assertAlmostEqual(evaluation['MAE_1_steps'], 0.094, places=2)
        self.assertAlmostEqual(evaluation['RMSE_3_steps'], 0.240, places=2)
        self.assertAlmostEqual(evaluation['MAE_3_steps'], 0.192, places=2)
        self.assertAlmostEqual(evaluation['RMSE'], 0.185, places=2)
        self.assertAlmostEqual(evaluation['MAE'], 0.144, places=2)

        # Test the cross validation
        reconstructor = PeriodicAverageReconstructor()
        cross_validation_results = reconstructor.cross_validate(timeseries, evaluate_steps=[1,3], fit_periodicity=24, evaluate_limit=100, evaluate_details=True)
        self.assertAlmostEqual(cross_validation_results['MAE_3_steps_avg'],  0.2293, places=2)
        self.assertAlmostEqual(cross_validation_results['MAE_3_steps_stdev'], 0.0313, places=2)

        # Test on Points as well
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        reconstructor = PeriodicAverageReconstructor()
        with self.assertRaises(Exception):
            reconstructor.fit(timeseries)

        timeseries = timeseries.resample(600)
        reconstructor.fit(timeseries)

        # Predict
        predicted_data = reconstructor.predict(timeseries, from_i=2, to_i=5)
        self.assertEqual(list(predicted_data.keys()), ['temperature'])
        self.assertEqual(len(predicted_data['temperature']), 4)
        self.assertAlmostEqual(predicted_data['temperature'][0], 23.71583333333333)
        self.assertAlmostEqual(predicted_data['temperature'][1], 23.750833333333325)
        self.assertAlmostEqual(predicted_data['temperature'][2], 23.784166666666664)
        self.assertAlmostEqual(predicted_data['temperature'][3], 23.7825)


    def test_ProphetReconstructor(self):
        try:
            import prophet
        except ImportError:
            print('Skipping Prophet tests as no prophet module installed')
            return

        # Get test data
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200).aggregate(3600)

        # Instantiate
        reconstructor = ProphetReconstructor()

        # Fit
        reconstructor.fit(timeseries)

        # Predict
        predicted_data = reconstructor.predict(timeseries, from_i=2, to_i=5)
        self.assertEqual(list(predicted_data.keys()), ['temperature_avg'])
        self.assertEqual(len(predicted_data['temperature_avg']), 4)
        self.assertAlmostEqual(predicted_data['temperature_avg'][0], 23.449303446403178)
        self.assertAlmostEqual(predicted_data['temperature_avg'][1], 23.45425065662797)
        self.assertAlmostEqual(predicted_data['temperature_avg'][2], 23.460265626928496)
        self.assertAlmostEqual(predicted_data['temperature_avg'][3], 23.46643974807204)

        # Can also predict on first and last items as it hasno window
        predicted_data = reconstructor.predict(timeseries, from_i=0, to_i=0)
        self.assertAlmostEqual(predicted_data['temperature_avg'][0], 23.439783345000002)
        predicted_data = reconstructor.predict(timeseries, from_i=len(timeseries)-1, to_i=len(timeseries)-1)
        self.assertAlmostEqual(predicted_data['temperature_avg'][0], 23.11027474748722)




import unittest
from ..datastructures import TimePoint, DataTimePoint , DataTimeSlot, TimeSeries
from ..operations import derivative, integral, diff, csum, min, max, avg, filter, select, mavg, normalize, merge, offset, rescale, slice, get
from ..operations import Operation

# Setup logging
from .. import logger
logger.setup()


class TestOpertions(unittest.TestCase):

    def test_base(self):

        operation_from_callable = Operation()
        self.assertEqual(operation_from_callable.__name__, 'operation')

        def operation_as_function(data):
            pass

        self.assertEqual(operation_as_function.__name__, 'operation_as_function')


    def test_diff_csum(self):

        # Test on empty, single point or variable resolution time series
        series = TimeSeries()
        with self.assertRaises(ValueError):
            diff(series)

        series = TimeSeries()
        series.append(DataTimePoint(t=0, data={'value':10}))
        with self.assertRaises(ValueError):
            diff(series)

        series = TimeSeries()
        series.append(DataTimePoint(t=0, data={'value':10}))
        series.append(DataTimePoint(t=1, data={'value':11}))
        series.append(DataTimePoint(t=8, data={'value':12}))
        with self.assertRaises(ValueError):
            diff(series)

        # Test data for the next tests
        series = TimeSeries()
        series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(60), data={'value':10}, data_loss=0.1))
        series.append(DataTimeSlot(start=TimePoint(60), end=TimePoint(120), data={'value':12}, data_loss=0.2))
        series.append(DataTimeSlot(start=TimePoint(120), end=TimePoint(180), data={'value':15}, data_loss=0.3))
        series.append(DataTimeSlot(start=TimePoint(180), end=TimePoint(240), data={'value':16}, data_loss=0.4))

        # Test standard, from the series
        diff_series = diff(series)
        self.assertEqual(len(diff_series), 3)
        self.assertEqual(diff_series[0].data['value_diff'],2)
        self.assertEqual(diff_series[1].data['value_diff'],3)
        self.assertEqual(diff_series[2].data['value_diff'],1)

        # Test data loss correctly carried forward by the diff
        self.assertEqual(diff_series[0].data_loss, 0.2)
        self.assertEqual(diff_series[1].data_loss, 0.3)
        self.assertEqual(diff_series[2].data_loss, 0.4)

        # Test csum as well
        diff_csum_series = csum(diff_series, offset=10)
        self.assertEqual(diff_csum_series[0].data['value_diff_csum'], 10)
        self.assertEqual(diff_csum_series[1].data['value_diff_csum'], 12)
        self.assertEqual(diff_csum_series[2].data['value_diff_csum'], 15)
        self.assertEqual(diff_csum_series[3].data['value_diff_csum'], 16)

        # Test data loss correctly carried forward by the csum
        self.assertEqual(diff_csum_series[0].data_loss, 0) # This is the slot created by the offset, has no data loss
        self.assertEqual(diff_csum_series[1].data_loss, 0.2)
        self.assertEqual(diff_csum_series[2].data_loss, 0.3)
        self.assertEqual(diff_csum_series[3].data_loss, 0.4)

        # Test standalone
        self.assertEqual(len(diff(series)), 3)

        # Multi data labels Test data (on points)
        series = TimeSeries()
        series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        series.append(DataTimePoint(t=60, data={'value':12, 'another_value': 65}))
        series.append(DataTimePoint(t=120, data={'value':6, 'another_value': 32}))
        series.append(DataTimePoint(t=180, data={'value':16, 'another_value': 10}))
        diff_series = series.diff()
        self.assertEqual(diff_series[0].data['value_diff'],2)
        self.assertEqual(diff_series[0].data['another_value_diff'],-10)
        self.assertEqual(diff_series[1].data['value_diff'],-6)
        self.assertEqual(diff_series[1].data['another_value_diff'],-33)

        # Test csum as well, with dict offset (on points)
        diff_csum_series = diff_series.csum(offset={'value_diff':10, 'another_value_diff':75})
        self.assertEqual(diff_csum_series[0].data['value_diff_csum'],10)
        self.assertEqual(diff_csum_series[0].data['another_value_diff_csum'],75)
        self.assertEqual(diff_csum_series[3].data['value_diff_csum'],16)
        self.assertEqual(diff_csum_series[3].data['another_value_diff_csum'],10)

        # Test on list data as well
        series = TimeSeries(DataTimePoint(t=60, data=[1.0,-4.0]),
                            DataTimePoint(t=120, data=[2.0,-2.0]),
                            DataTimePoint(t=180, data=[4.0,-1.0]))
        self.assertEqual(len(series.diff()),2)
        self.assertEqual(series.diff()[0].data, [1.0, 2.0])
        self.assertEqual(series.diff()[1].data, [2.0, 1.0])
        self.assertEqual(len(series.csum()),3)
        self.assertEqual(series.csum()[0].data, [1.0, -4.0])
        self.assertEqual(series.csum()[1].data, [3.0, -6.0])
        self.assertEqual(series.csum()[2].data, [7.0, -7.0])


    def test_derivative_integral(self):

        # Test on empty, single point or variable resolution time serie
        with self.assertRaises(ValueError):
            series = TimeSeries()
            derivative(series)

        with self.assertRaises(ValueError):
            series = TimeSeries()
            series.append(DataTimePoint(t=0, data={'value':10}))
            derivative(series)

        # Test data
        series = TimeSeries()
        series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(1), data={'value':10}, data_loss=0.1))
        series.append(DataTimeSlot(start=TimePoint(1), end=TimePoint(2), data={'value':12}, data_loss=0.2))
        series.append(DataTimeSlot(start=TimePoint(2), end=TimePoint(3), data={'value':15}, data_loss=0.3))
        series.append(DataTimeSlot(start=TimePoint(3), end=TimePoint(4), data={'value':16}, data_loss=0.4))

        # Test standard derivative behavior, from the series
        derivative_series = derivative(series)
        self.assertEqual(len(derivative_series), 4)
        self.assertAlmostEqual(derivative_series[0].data['value_derivative'],2)     # 0
        self.assertAlmostEqual(derivative_series[1].data['value_derivative'],2.5)   # 2
        self.assertAlmostEqual(derivative_series[2].data['value_derivative'],2)     # 5
        self.assertAlmostEqual(derivative_series[3].data['value_derivative'],1.0)   # 6

        # Test data loss correctly carried forward by the derivative
        self.assertEqual(derivative_series[0].data_loss, 0.1)
        self.assertEqual(derivative_series[1].data_loss, 0.2)
        self.assertEqual(derivative_series[2].data_loss, 0.3)
        self.assertEqual(derivative_series[3].data_loss, 0.4)

        # Test integral as well
        derivative_integral_series = integral(derivative_series, c=10)
        self.assertEqual(derivative_integral_series[0].data['value_derivative_integral'], 10)
        self.assertEqual(derivative_integral_series[1].data['value_derivative_integral'], 12)
        self.assertEqual(derivative_integral_series[2].data['value_derivative_integral'], 15)
        self.assertEqual(derivative_integral_series[3].data['value_derivative_integral'], 16)

        # Test data loss correctly carried forward by the integral
        self.assertEqual(derivative_integral_series[0].data_loss, 0.1)
        self.assertEqual(derivative_integral_series[1].data_loss, 0.2)
        self.assertEqual(derivative_integral_series[2].data_loss, 0.3)
        self.assertEqual(derivative_integral_series[3].data_loss, 0.4)

        # Test data
        series = TimeSeries()
        series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(10), data={'value':10}))
        series.append(DataTimeSlot(start=TimePoint(10), end=TimePoint(20), data={'value':12}))
        series.append(DataTimeSlot(start=TimePoint(20), end=TimePoint(30), data={'value':15}))
        series.append(DataTimeSlot(start=TimePoint(30), end=TimePoint(40), data={'value':16}))

        # Test standard derivative behavior, from the series
        derivative_series = derivative(series)
        self.assertEqual(len(derivative_series), 4)
        self.assertAlmostEqual(derivative_series[0].data['value_derivative'],0.2)
        self.assertAlmostEqual(derivative_series[1].data['value_derivative'],0.25)
        self.assertAlmostEqual(derivative_series[2].data['value_derivative'],0.2)
        self.assertAlmostEqual(derivative_series[3].data['value_derivative'],0.1)

        # Multi data labels Test data
        series = TimeSeries()
        series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        series.append(DataTimePoint(t=10, data={'value':12, 'another_value': 65}))
        series.append(DataTimePoint(t=20, data={'value':6, 'another_value': 32}))
        series.append(DataTimePoint(t=30, data={'value':16, 'another_value': 10}))
        derivative_series = series.derivative()
        self.assertAlmostEqual(derivative_series[0].data['value_derivative'],0.2)
        self.assertAlmostEqual(derivative_series[0].data['another_value_derivative'],-1.0)
        self.assertAlmostEqual(derivative_series[1].data['value_derivative'],-0.2)
        self.assertAlmostEqual(derivative_series[1].data['another_value_derivative'],-2.15)

        # Test derivative-integral identity with the data
        series = TimeSeries()
        series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(1), data={'value':0}))
        series.append(DataTimeSlot(start=TimePoint(1), end=TimePoint(2), data={'value':6}))
        series.append(DataTimeSlot(start=TimePoint(2), end=TimePoint(3), data={'value':14}))
        series.append(DataTimeSlot(start=TimePoint(3), end=TimePoint(4), data={'value':16}))
        series.append(DataTimeSlot(start=TimePoint(4), end=TimePoint(5), data={'value':20}))

        derivative_series = series.derivative()
        derivative_integral_series = derivative_series.integral()

        self.assertAlmostEqual(derivative_integral_series[0].data['value_derivative_integral'],0)
        self.assertAlmostEqual(derivative_integral_series[1].data['value_derivative_integral'],6)
        self.assertAlmostEqual(derivative_integral_series[2].data['value_derivative_integral'],14)
        self.assertAlmostEqual(derivative_integral_series[3].data['value_derivative_integral'],16)
        self.assertAlmostEqual(derivative_integral_series[4].data['value_derivative_integral'],20)

        # Variable sampling rate
        series = TimeSeries()
        series.append(DataTimePoint(t=0,  data={'value':0}))
        series.append(DataTimePoint(t=1,  data={'value':0.5}))
        series.append(DataTimePoint(t=2,  data={'value':1}))
        series.append(DataTimePoint(t=4,  data={'value':2.5}))
        series.append(DataTimePoint(t=5,  data={'value':3.5}))

        derivative_series = series.derivative()
        self.assertAlmostEqual(derivative_series[0].data['value_derivative'], 0.5)
        self.assertAlmostEqual(derivative_series[1].data['value_derivative'], 0.5)
        self.assertAlmostEqual(derivative_series[2].data['value_derivative'], 0.625)
        self.assertAlmostEqual(derivative_series[3].data['value_derivative'], 0.875)
        self.assertAlmostEqual(derivative_series[4].data['value_derivative'], 1.0)

        derivative_integral_series = derivative_series.integral()

        self.assertAlmostEqual(derivative_integral_series[0].data['value_derivative_integral'],0)
        self.assertAlmostEqual(derivative_integral_series[1].data['value_derivative_integral'],0.5)
        self.assertAlmostEqual(derivative_integral_series[2].data['value_derivative_integral'],1)
        self.assertAlmostEqual(derivative_integral_series[3].data['value_derivative_integral'],2.5)
        self.assertAlmostEqual(derivative_integral_series[4].data['value_derivative_integral'],3.5)


    def test_normalize(self):
        series =  TimeSeries(DataTimePoint(t=60, data={'a':2, 'b':6}, data_loss=0.1),
                             DataTimePoint(t=120, data={'a':4, 'b':9}, data_loss=0.2),
                             DataTimePoint(t=180, data={'a':8, 'b':3}, data_loss=0.3))

        normalized_series = normalize(series)

        self.assertEqual(normalized_series[0].data['a'],0)
        self.assertEqual(normalized_series[2].data['b'],0)

        self.assertEqual(normalized_series[2].data['a'],1)
        self.assertEqual(normalized_series[1].data['b'],1)

        # Test data loss correctly carried forward by the operation
        self.assertEqual(normalized_series[0].data_loss, 0.1)
        self.assertEqual(normalized_series[1].data_loss, 0.2)
        self.assertEqual(normalized_series[2].data_loss, 0.3)

        # Test normalize with respect to another range (other than 0-1):
        custom_normalized_series = normalize(series, [0.5,1.5])

        self.assertEqual(custom_normalized_series[0].data['a'],0.5)
        self.assertAlmostEqual(custom_normalized_series[1].data['a'],0.83333333)
        self.assertEqual(custom_normalized_series[2].data['a'],1.5)

        # Test on list data as well
        series = TimeSeries(DataTimePoint(t=60, data=[1.0,-3.0]),
                            DataTimePoint(t=120, data=[2.0,-2.0]),
                            DataTimePoint(t=180, data=[3.0,-1.0]))
        self.assertEqual(series.normalize()[0].data, [0.0,0.0])
        self.assertEqual(series.normalize()[1].data, [0.5,0.5])
        self.assertEqual(series.normalize()[2].data, [1.0,1.0])


    def test_rescale(self):
        series = TimeSeries(DataTimePoint(t=60, data={'a':2, 'b':6}, data_loss=0.1),
                            DataTimePoint(t=120, data={'a':4, 'b':9}, data_loss=0.2),
                            DataTimePoint(t=180, data={'a':8, 'b':3}, data_loss=0.3))

        # Rescale with single value
        rescaled_series = rescale(series, 2)
        self.assertEqual(len(rescaled_series),3)
        self.assertEqual(rescaled_series[0].data['a'],4)
        self.assertEqual(rescaled_series[0].data['b'],12)
        self.assertEqual(rescaled_series[1].data['a'],8)
        self.assertEqual(rescaled_series[1].data['b'],18)
        self.assertEqual(rescaled_series[2].data['a'],16)
        self.assertEqual(rescaled_series[2].data['b'],6)

        # Rescale with specific value
        rescaled_series = rescale(series, {'b':2})
        self.assertEqual(len(rescaled_series),3)
        self.assertEqual(rescaled_series[0].data['a'],2)
        self.assertEqual(rescaled_series[0].data['b'],12)
        self.assertEqual(rescaled_series[1].data['a'],4)
        self.assertEqual(rescaled_series[1].data['b'],18)
        self.assertEqual(rescaled_series[2].data['a'],8)
        self.assertEqual(rescaled_series[2].data['b'],6)

        # Test data loss correctly carried forward by the operation
        self.assertEqual(rescaled_series[0].data_loss, 0.1)
        self.assertEqual(rescaled_series[1].data_loss, 0.2)
        self.assertEqual(rescaled_series[2].data_loss, 0.3)

        # Test on list data as well
        series = TimeSeries(DataTimePoint(t=60, data=[1.0,-3.0]),
                            DataTimePoint(t=120, data=[2.0,-2.0]),
                            DataTimePoint(t=180, data=[3.0,-1.0]))
        self.assertEqual(series.rescale(10)[0].data, [10.0,-30.0])
        self.assertEqual(series.rescale(10)[1].data, [20.0,-20.0])
        self.assertEqual(series.rescale(10)[2].data, [30.0,-10.0])


    def test_offset(self):
        series = TimeSeries(DataTimePoint(t=60, data={'a':2, 'b':6}, data_loss=0.1),
                            DataTimePoint(t=120, data={'a':4, 'b':9}, data_loss=0.2),
                            DataTimePoint(t=180, data={'a':8, 'b':3}, data_loss=0.3))

        # Offset with single value
        offsetted_series = offset(series, 10)
        self.assertEqual(len(offsetted_series),3)
        self.assertEqual(offsetted_series[0].data['a'],12)
        self.assertEqual(offsetted_series[0].data['b'],16)
        self.assertEqual(offsetted_series[1].data['a'],14)
        self.assertEqual(offsetted_series[1].data['b'],19)
        self.assertEqual(offsetted_series[2].data['a'],18)
        self.assertEqual(offsetted_series[2].data['b'],13)

        # Offset with specific value
        offsetted_series = offset(series, {'b':-3})
        self.assertEqual(len(offsetted_series),3)
        self.assertEqual(offsetted_series[0].data['a'],2)
        self.assertEqual(offsetted_series[0].data['b'],3)
        self.assertEqual(offsetted_series[1].data['a'],4)
        self.assertEqual(offsetted_series[1].data['b'],6)
        self.assertEqual(offsetted_series[2].data['a'],8)
        self.assertEqual(offsetted_series[2].data['b'],0)

        # Test data loss correctly carried forward by the operation
        self.assertEqual(offsetted_series[0].data_loss, 0.1)
        self.assertEqual(offsetted_series[1].data_loss, 0.2)
        self.assertEqual(offsetted_series[2].data_loss, 0.3)

        # Test on list data as well
        series = TimeSeries(DataTimePoint(t=60, data=[1.0,-3.0]),
                            DataTimePoint(t=120, data=[2.0,-2.0]),
                            DataTimePoint(t=180, data=[3.0,-1.0]))
        self.assertEqual(series.offset(10)[0].data, [11.0,7.0])
        self.assertEqual(series.offset(10)[1].data, [12.0,8.0])
        self.assertEqual(series.offset(10)[2].data, [13.0,9.0])


    def test_mavg(self):

        series = TimeSeries()
        series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(1), data={'value':2}, data_loss=0.1))
        series.append(DataTimeSlot(start=TimePoint(1), end=TimePoint(2), data={'value':4}, data_loss=0.2))
        series.append(DataTimeSlot(start=TimePoint(2), end=TimePoint(3), data={'value':8}, data_loss=0.3))
        series.append(DataTimeSlot(start=TimePoint(3), end=TimePoint(4), data={'value':16}, data_loss=0.4))

        mavg_series = mavg(series, 2)

        # Check len and timestamps (the first is skipped due to moving average window of two)
        self.assertEqual(len(mavg_series),3)
        self.assertEqual(mavg_series[0].t, 1)
        self.assertEqual(mavg_series[1].t, 2)
        self.assertEqual(mavg_series[2].t, 3)

        # Check data
        self.assertEqual(mavg_series[0].data['value_mavg_2'], 3)
        self.assertEqual(mavg_series[1].data['value_mavg_2'], 6)
        self.assertEqual(mavg_series[2].data['value_mavg_2'], 12)

        # Check data loss removed (does not make sense in a moving average logic)
        self.assertEqual(mavg_series[0].data_loss, None)
        self.assertEqual(mavg_series[1].data_loss, None)
        self.assertEqual(mavg_series[2].data_loss, None)

        # Test on list data as well
        series = TimeSeries(DataTimePoint(t=60, data=[1.0,-3.0]),
                            DataTimePoint(t=120, data=[2.0,-2.0]),
                            DataTimePoint(t=180, data=[3.0,-1.0]))
        self.assertEqual(len(series.mavg(2)),2)
        self.assertEqual(series.mavg(2)[0].data, [1.5, -2.5])
        self.assertEqual(series.mavg(2)[1].data, [2.5, -1.5])


    def test_min_max_avg(self):

        # Test data
        series = TimeSeries()
        series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(60), data={'value':10}))
        series.append(DataTimeSlot(start=TimePoint(60), end=TimePoint(120), data={'value':12}))
        series.append(DataTimeSlot(start=TimePoint(120), end=TimePoint(180), data={'value':6}))
        series.append(DataTimeSlot(start=TimePoint(180), end=TimePoint(240), data={'value':16}))

        # Test standalone
        self.assertEqual(min(series), {'value':6})
        self.assertEqual(max(series), {'value':16})
        self.assertEqual(avg(series), {'value':11})

        # Test from the series
        self.assertEqual(series.min(), {'value':6})
        self.assertEqual(series.max(), {'value':16})
        self.assertEqual(series.avg(), {'value':11})

        # Multi data labels Test data
        series = TimeSeries()
        series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        series.append(DataTimePoint(t=60, data={'value':12, 'another_value': 65}))
        series.append(DataTimePoint(t=120, data={'value':6, 'another_value': 32}))
        series.append(DataTimePoint(t=180, data={'value':16, 'another_value': 10}))

        self.assertEqual(series.min(), {'value': 6, 'another_value': 10})
        self.assertEqual(series.min(data_label='value'), 6)
        self.assertEqual(series.min(data_label='another_value'), 10)

        self.assertEqual(series.max(), {'value': 16, 'another_value': 75})
        self.assertEqual(series.max(data_label='value'), 16)
        self.assertEqual(series.max(data_label='another_value'), 75)

        self.assertEqual(series.avg(), {'value': 11, 'another_value': 45.5})
        self.assertEqual(series.avg(data_label='value'), 11)
        self.assertEqual(series.avg(data_label='another_value'), 45.5)

        # Test on list data as well
        series = TimeSeries(DataTimePoint(t=60, data=[5.8,29.6]), DataTimePoint(t=120, data=[7.8,18.6]))
        self.assertEqual(series.max(), {'0': 7.8, '1': 29.6})
        self.assertEqual(series.min(), {'0': 5.8, '1': 18.6})
        self.assertEqual(series.avg(), {'0': 6.8, '1': 24.1})


    def test_avg_weighted(self):

        series1 = TimeSeries()
        series2 = TimeSeries()

        point = DataTimePoint(t=-3, data={'value':3})
        point.weight=0
        series1.append(point)
        series2.append(point)

        point = DataTimePoint(t=-2, data={'value':2})
        point.weight=0.125
        series1.append(point)

        point = DataTimePoint(t=-1, data={'value':1})
        point.weight=0.25
        series1.append(point)
        series2.append(point)

        point = DataTimePoint(t=0, data={'value':0})
        point.weight=0.25
        series1.append(point)
        series2.append(point)

        point = DataTimePoint(t=1, data={'value':1})
        point.weight=0.25
        series1.append(point)
        series2.append(point)

        point = DataTimePoint(t=2, data={'value':2})
        point.weight=0.125
        series1.append(point)
        series2.append(point)

        # [DEBUG] timeseria.operations: Point @ 1969-12-31 23:59:57+00:00, weight: 0
        # [DEBUG] timeseria.operations: Point @ 1969-12-31 23:59:58+00:00, weight: 0.125
        # [DEBUG] timeseria.operations: Point @ 1969-12-31 23:59:59+00:00, weight: 0.25
        # [DEBUG] timeseria.operations: Point @ 1970-01-01 00:00:00+00:00, weight: 0.25
        # [DEBUG] timeseria.operations: Point @ 1970-01-01 00:00:01+00:00, weight: 0.25
        # [DEBUG] timeseria.operations: Point @ 1970-01-01 00:00:02+00:00, weight: 0.125

        self.assertEqual(avg(series1), {'value':1})
        self.assertEqual(avg(series2), {'value':0.8571428571428571})


    def test_get(self):
        series = TimeSeries()
        series.append(DataTimePoint(t=0, data={'value':0}))
        series.append(DataTimePoint(t=60, data={'value':60}))
        series.append(DataTimePoint(t=120, data={'value':120}))
        series.append(DataTimePoint(t=180, data={'value':180}))

        # By int (index)
        self.assertEqual(get(series, 0), series[0])
        self.assertEqual(get(series, 1), series[1])
        self.assertEqual(get(series, -1), series[3])

        # By float (epoch timestamp)
        self.assertEqual(get(series, at_t=float(0)), series[0])
        self.assertEqual(get(series, at_t=float(60)), series[1])

        # By datetime
        self.assertEqual(get(series, at_dt=series[0].dt), series[0])
        self.assertEqual(get(series, at_dt=series[1].dt), series[1])


    def test_filter(self):

        # Test data
        series =  TimeSeries(DataTimePoint(t=60, data={'a':1, 'b':2, 'c':3}),
                             DataTimePoint(t=120, data={'a':2, 'b':4, 'c':4}),
                             DataTimePoint(t=180, data={'a':3, 'b':8, 'c':5}))

        # Test get item by string key (filtering on data labels)
        self.assertEqual(filter(series, 'b')[0].data, {'b': 2})

        # Test get item by string key (filtering on data labels), from the series
        self.assertEqual(series.filter('a').data_labels(), ['a'])
        self.assertEqual(len(series.filter('a')), 3 )
        self.assertEqual(series.filter('a')[0].data, {'a': 1})
        self.assertEqual(series.filter('a')[1].data, {'a': 2})
        self.assertEqual(series.filter('a')[2].data, {'a': 3})


        # Test that we haven't modified the original series
        self.assertEqual(series.data_labels(), ['a', 'b', 'c'])
        self.assertEqual(series[1].data['b'], 4)

        # Test for multiple filtering data labels
        self.assertEqual(set(series.filter('a', 'c').data_labels()), {'a', 'c'})
        self.assertEqual(len(series.filter('a', 'c')), 3 )
        self.assertEqual(series.filter('a', 'c')[0].data, {'a': 1,'c':3})
        self.assertEqual(series.filter('a', 'c')[1].data, {'a': 2, 'c':4})
        self.assertEqual(series.filter('a', 'c')[2].data, {'a': 3, 'c':5})

        # Test error when filtering for non key-value  data
        series =  TimeSeries(DataTimePoint(t=60, data=[1,2]),
                             DataTimePoint(t=120, data=[3,4]),
                             DataTimePoint(t=180, data=[5,6]))
        with self.assertRaises(TypeError):
            series.filter('0')


    def test_slice(self):

        series = TimeSeries(TimePoint(t=60),TimePoint(t=120),TimePoint(t=130),TimePoint(t=140),TimePoint(t=150),TimePoint(t=160))

        # Test start/end slicing
        self.assertEqual(len(slice(series, from_t=1.0, to_t=140)),3)

        # Test start/end slicing from the series
        self.assertEqual(len(series.slice(from_t=float(1),to_t=float(140))),3)
        self.assertEqual(len(series.slice(from_t=float(61), to_t=float(140))),2)
        self.assertEqual(len(series.slice(from_t=float(61))),5)
        self.assertEqual(len(series.slice(to_t=float(61))),1)

        # Test no keyword arguments
        self.assertEqual(len(series.slice(from_t=float(61),to_t=float(140))),2)

        # Test with slots
        slot_series = TimeSeries(DataTimeSlot(start=TimePoint(t=60), end=TimePoint(t=120),data=1),
                                 DataTimeSlot(start=TimePoint(t=120), end=TimePoint(t=180),data=1),
                                 DataTimeSlot(start=TimePoint(t=180), end=TimePoint(t=240),data=1),
                                 DataTimeSlot(start=TimePoint(t=240), end=TimePoint(t=300),data=1))
        self.assertEqual(len(slice(slot_series,from_t=float(60),to_t=float(120))),1)
        self.assertEqual(len(slice(slot_series,from_t=float(60),to_t=float(121))),1)
        self.assertEqual(len(slice(slot_series,from_t=float(60),to_t=float(180))),2)

        # Check also time zone
        slot_series.change_tz('Europe/Rome')
        self.assertEqual(slot_series.slice(from_t=float(60),to_t=float(60)).tz, None)
        self.assertEqual(str(slot_series.slice(from_t=float(60),to_t=float(120)).tz), 'Europe/Rome')


    def test_select(self):

        # Test data
        series = TimeSeries()
        series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        series.append(DataTimePoint(t=60, data={'value':12, 'another_value': 65}))
        series.append(DataTimePoint(t=120, data={'value':6, 'another_value': 32}))
        series.append(DataTimePoint(t=180, data={'value':16, 'another_value': 10}))

        # Basic select test
        self.assertEqual(select(series, query='another_value=65')[0].t, 60)
        self.assertEqual(series.select('"another_value"=65')[0].t, 60)


    def test_merge(self):

        # Test data
        series1 = TimeSeries()
        series1.append(DataTimePoint(t=0, data={'value':10}, data_indexes={'data_loss':0.1, 'index_a':0.1}))
        series1.append(DataTimePoint(t=60, data={'value':12}, data_indexes={'data_loss':0.2, 'index_a':0.2}))
        series1.append(DataTimePoint(t=120, data={'value':6}, data_indexes={'data_loss':0.3, 'index_a':0.3}))
        series1.append(DataTimePoint(t=180, data={'value':16}, data_indexes={'data_loss':0.4, 'index_a':0.4}))
        series1.append(DataTimePoint(t=240, data={'value':20}, data_indexes={'data_loss':0.5, 'index_a':0.5}))

        series2 = TimeSeries()
        series2.append(DataTimePoint(t=0, data={'another_value': 75}, data_indexes={'data_loss':0.01, 'index_b':0.01}))
        series2.append(DataTimePoint(t=60, data={'another_value': 65}, data_indexes={'data_loss':0.02, 'index_b':0.02}))
        series2.append(DataTimePoint(t=120, data={'another_value': 32}, data_indexes={'data_loss':0.03, 'index_b':0.03}))
        series2.append(DataTimePoint(t=180, data={'another_value': 10}, data_indexes={'data_loss':None, 'index_b':0.04}))
        series2.append(DataTimePoint(t=240, data={'another_value': 7}, data_indexes={'data_loss':None, 'index_b':0.05}))

        series3 = TimeSeries()
        series3.append(DataTimePoint(t=0, data={'another_value': 10}, data_indexes={'data_loss':0.01, 'index_b':0.01}))
        series3.append(DataTimePoint(t=60, data={'another_value': 10}, data_indexes={'data_loss':0.02, 'index_b':0.02}))
        series3.append(DataTimePoint(t=120, data={'another_value': 10}, data_indexes={'data_loss':0.03, 'index_b':0.03}))
        series3.append(DataTimePoint(t=180, data={'another_value': 10}, data_indexes={'data_loss':None, 'index_b':0.04}))
        series3.append(DataTimePoint(t=240, data={'another_value': 10}, data_indexes={'data_loss':None, 'index_b':0.05}))

        series4 = TimeSeries()
        series4.append(DataTimePoint(t=0, data={'value': 10}))
        series4.append(DataTimePoint(t=60, data={'value': 11}))
        series4.append(DataTimePoint(t=120, data={'value': 12}))
        series4.append(DataTimePoint(t=180, data={'value': 13}))
        series4.append(DataTimePoint(t=240, data={'value': 14}))

        # Basic merge
        merged = merge(series1,series2)
        self.assertEqual(len(merged), 5)
        self.assertEqual(merged[0].t, 0)
        self.assertEqual(merged[-1].t, 240)
        self.assertEqual(merged[0].data['value'], 10)
        self.assertEqual(merged[0].data['another_value'], 75)
        self.assertEqual(merged[-1].data['value'], 20)
        self.assertEqual(merged[-1].data['another_value'], 7)

        # Subset merge
        merged = merge(series1[2:4],series2)
        self.assertEqual(len(merged), 2)
        self.assertEqual(merged[0].t, 120)
        self.assertEqual(merged[-1].t, 180)
        self.assertEqual(merged[0].data['value'], 6)
        self.assertEqual(merged[0].data['another_value'], 32)
        self.assertEqual(merged[-1].data['value'], 16)
        self.assertEqual(merged[-1].data['another_value'], 10)

        # Data math
        merged = merge(series2,series3)
        self.assertEqual(len(merged), 5)
        self.assertEqual(merged[0].t, 0)
        self.assertEqual(merged[-1].t, 240)
        self.assertEqual(merged[0].data['another_value'], (75+10)/2)
        self.assertEqual(merged[-1].data['another_value'], (7+10)/2)

        # Data indexes math
        merged = merge(series1,series2)
        self.assertEqual(merged[0].data_indexes, {'data_loss': (0.1+0.01)/2, 'index_a': 0.1, 'index_b': 0.01})
        self.assertEqual(merged[1].data_indexes, {'data_loss': (0.2+0.02)/2, 'index_a': 0.2, 'index_b': 0.02})
        self.assertEqual(merged[2].data_indexes, {'data_loss': (0.3+0.03)/2, 'index_a': 0.3, 'index_b': 0.03})
        self.assertEqual(merged[3].data_indexes, {'data_loss': 0.4, 'index_a': 0.4, 'index_b': 0.04})
        self.assertEqual(merged[4].data_indexes, {'data_loss': 0.5, 'index_a': 0.5, 'index_b': 0.05})

        # Merge different data types preserving them
        merged = merge(series3,series4)
        self.assertTrue(isinstance(merged[0].data['value'], int))
        self.assertTrue(isinstance(merged[1].data['value'], int))
        self.assertTrue(isinstance(merged[2].data['value'], int))
        self.assertTrue(isinstance(merged[3].data['value'], int))
        self.assertTrue(isinstance(merged[4].data['value'], int))




# -*- coding: utf-8 -*-
"""Units, including the TimeUnit, which fully supports calendar arithmetic."""

from .utils import is_numerical
from propertime import TimeSpan, Time

# Setup logging
import logging
logger = logging.getLogger(__name__)


class Unit(object):
    """A generic unit.

       Args:
           value: the unit value.
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return str(self.value)

    def __add__(self, other):

        # Sum with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value + other

        # Sum with another Unit
        if self.__class__ == other.__class__:
            return self.value + other.value

        # Sum with an objects with coordinates (Points, Slots)
        try:
            if len(other.coordinates) > 1:
                raise NotImplementedError('Cannot sum multidimensional objects')
            return other.__class__(self.value + other.coordinates[0])
        except:
            raise NotImplementedError('Don\'t know how to add Units with {}'.format(other.__class__.__name__))

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):

        # Subtract with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value - other

        # Subtract with another Unit
        if self.__class__ == other.__class__:
            return self.value - other.value

        # Sum with an objects with coordinates (Points, Slots)
        try:
            if len(other.coordinates) > 1:
                raise NotImplementedError('Cannot subtract multidimensional objects')
            return other.__class__(self.value - other.coordinates[0])
        except:
            raise NotImplementedError('Don\'t know how to subtract Units with {}'.format(other.__class__.__name__))

    def __rsub__(self, other):
        return -self.__sub__(other)

    def __truediv__(self, other):

        # Divide with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value / other
        else:
            if is_numerical(self.value):
                raise NotImplementedError('Division with "{}" is not implemented'.format(other.__class__.__name__))
            else:
                raise NotImplementedError('Cannot divide multidimensional objects')

    def __rtruediv__(self, other):
        return ( 1 / self.__truediv__(other))

    def __mul__(self, other):

        # Multiply with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value * other
        else:
            if is_numerical(self.value):
                raise NotImplementedError('Division with "{}" is not implemented'.format(other.__class__.__name__))
            else:
                raise NotImplementedError('Cannot divide multidimensional objects')

    def __rmul__(self, other):
        return self.__mul__(other)

    def __eq__(self,other):
        if self is other:
            return True
        elif is_numerical(other):
            return self.value == other
        else:
            try:
                return self.value == other.value
            except AttributeError:
                return False



class TimeUnit(TimeSpan, Unit):
    """A unit that can have both fixed (physical) or variable (calendar) time length.
    It can handle precision up to the microsecond and can be added and subtracted with numerical
    values, Time and datetime objects, and other TimeUnits.

    Can be initialized both using a numerical value, a string representation, or by explicitly setting
    years, months, weeks, days, hours, minutes and seconds (including sub-second precision).
    In the string representation, the mapping is as follows:

        * ``'Y': 'years'``
        * ``'M': 'months'``
        * ``'W': 'weeks'``
        * ``'D': 'days'``
        * ``'h': 'hours'``
        * ``'m': 'minutes'``
        * ``'s': 'seconds'``

    For example, to create a time unit of one hour, the following three are equivalent, where the
    first one uses the numerical value, the second the string representation, and the third explicitly
    sets the time component (hours in this case): ``TimeUnit('1h')``, ``TimeUnit(hours=1)``, or ``TimeUnit(3600)``.
    Not all time units can be initialized using the numerical value, in particular calendar time units which can
    have variable duration: a time unit of one day, or ``TimeUnit('1d')``, can last for 23, 24 or 24 hours depending
    on DST changes. On the contrary, a ``TimeUnit('24h')`` will always last 24 hours and can be initialized as
    ``TimeUnit(86400)`` as well.

    Args:
        value: the time unit value, either as seconds (int or float) or as string representation according to the mapping above.
        years: the time unit years component (int).
        weeks: the time unit weeks component (int).
        months: the time unit weeks component (int).
        days: the time unit days component (int).
        hours: the time unit hours component (int).
        minutes: the time unit minutes component (int).
        seconds: the time unit seconds component, including sub-second precision (int, float).
    """

    # Support value-based init
    def __init__(self, *args, **kwargs):
        if args and (isinstance(args[0], int) or isinstance(args[0], float)):
            args = list(args)
            args[0] = '{}s'.format(args[0])
        if kwargs and 'value' in kwargs and (isinstance(kwargs['seconds'], int) or isinstance(kwargs['seconds'], float)) :
            kwargs['seconds'] = kwargs.pop('value')
        super(TimeUnit, self).__init__(*args, **kwargs)

    # Support for TimePoints and this TimeUnit class
    def __radd__(self, other):

        # Import here to avoid cyclic imports
        from .datastructures import TimePoint
        if isinstance(other, TimePoint):
            return TimePoint(dt=self.shift(other.dt, times=1))
        elif isinstance(other, self.__class__):
            return TimeUnit(years        = self.years + other.years,
                            months       = self.months + other.months,
                            weeks        = self.weeks + other.weeks,
                            days         = self.days + other.days,
                            hours        = self.hours + other.hours,
                            minutes      = self.minutes + other.minutes,
                            seconds      = self.seconds + other.seconds)
        else:
            return super().__radd__(other)

    def __rsub__(self, other):

        # Import here to avoid cyclic imports
        from .datastructures import TimePoint
        if isinstance(other, TimePoint):
            return TimePoint(dt=self.shift(other.dt, times=-1))
        else:
            return super().__rsub__(other)

    # For the doc strings
    def as_seconds(self, starting_at=None):
        """The length (duration) of the time unit, in seconds.

            Args:
                starting_at(Time, datetime): the starting point with respect to which compute the duration. Required for
                                             all calendar time units (involving years, months, weeks and days). Defaults to None.

            Returns:
                float: the duration of the unit in seconds.
        """
        return super().as_seconds(starting_at)

    def ceil(self, time):
        """Ceil a Time or datetime object according to this time unit.

            Args:
                time(Time, datetime): Time or datetime object to be ceiled.

            Returns:
                Time or datetime: the ceiled Time or datetime object.
        """
        return super().ceil(time)

    def floor(self, time):
        """Floor a Time or datetime object according to this time unit.

            Args:
                time(Time, datetime): the Time or datetime object to be floored.

            Returns:
                Time or datetime: the floored Time or datetime object.
        """
        return super().floor(time)

    def round(self, time, how='half'):
        """Round a Time or datetime object according to this time unit.

            Args:
                time(Time, datetime): the Time or datetime object to be rounded.

            Returns:
                Time or datetime: the rounded Time or datetime object.
        """
        return super().round(time, how)

    def shift(self, time, times=1):
        """Shift a Time or datetime object n times this time unit.

            Args:
                time(Time, datetime): the Time or datetime object to be shifted.
                times(int): how many times to perform the shift. Defaults to 1.

            Returns:
                Time or datetime: the shifted Time or datetime object.
        """
        return super().shift(time, times)




# -*- coding: utf-8 -*-
"""Interpolators."""

# Setup logging
import logging
logger = logging.getLogger(__name__)


class Interpolator():
    """A generic interpolator.

        Args:
            series(TimeSeries): the series on which to initialize the interpolator.
    """

    def __init__(self, series):
        logger.debug('Initialized %s interpolator on series "&s"'.format(self.__class__.__name__, series))
        self.series = series

    def evaluate(self, at, prev_i, next_i):
        """Evaluate the interpolator (compute the value).

            Args:
                at(float): the time value (as epoch seconds) at which to evaluate the interpolator.
                prev_i(int): the previous position of the item in the series.
                next_i(int): the next position of the item in the series.
        """
        raise NotImplementedError('This interpolator is not implemented')


class LinearInterpolator(Interpolator):
    """A linear interpolator.

        Args:
            series(TimeSeries): the series on which to initialize the interpolator.
    """

    def evaluate(self, at, prev_i=None, next_i=None):

        if prev_i is not None and next_i is None:
            raise ValueError('If you provide the prev_i you also need to provide the next_i')
        if next_i is not None  and prev_i is None:
            raise ValueError('If you provide the next_i you also need to provide the prev_i')

        if prev_i is None and next_i is None:
            logger.warning('You are not providing the prev_i and next_i, this will cause the interpolator to look for them and likely introduce a slow down')

            # Search for the prev_i:
            for i,item in enumerate(self.series):
                if item.t > at:
                    prev_i = i-1
                    next_i = i
                    break

            if next_i is None:
                raise ValueError('Trying to interpolate a point outside the series: prev_i="{}", next_i="{}", at="{}", series="{}"'.format(prev_i, next_i, at, self.series))

        else:
            if not self.series[prev_i].t < at < self.series[next_i].t:
                raise ValueError('Trying to interpolate a point outside the prev_i / next_i interval')

        # Shortcuts
        prev_point = self.series[prev_i]
        next_point = self.series[next_i]

        # Compute the increment with respect to the interpolation
        interpolated_data = self.series[0].data.__class__()

        # Example:
        # Indexes  0 1 2 3 4 5 6 7
        # Coords:  1 2 3 4 5 6 7 8
        # Values   1 2 x x x 6 7 8
        # prev_i = 1, next_i = 5
        # at = 3 (i=2)
        # interpolation: 4

        for label in self.series.data_labels():

            # Compute the "growth" ratio
            diff = next_point.data[label] - prev_point.data[label]
            delta_t = next_point.t - prev_point.t
            ratio = diff / delta_t

            # Compute the value of the data for the new point
            interpolated_data[label] = prev_point.data[label] + ((at-prev_point.t)*ratio)


        return interpolated_data


class UniformInterpolator(Interpolator):
    """A unform interpolator.

        Args:
            series(TimeSeries): the series on which to initialize the interpolator.
    """
    def evaluate(self, at, prev_i=None, next_i=None):

        if prev_i and next_i is None:
            raise ValueError('If you provide the prev_i you also need to provide the next_i')
        if next_i and prev_i is None:
            raise ValueError('If you provide the next_i you also need to provide the prev_i')

        if not prev_i:
            logger.warning('You are not providing the prev_i and next_i, this will cause the interpolator to look for them and likely introduce a slow down')

            # Search for the prev_i:
            for i,item in enumerate(self.series):
                if item.t > at:
                    prev_i = i-1
                    next_i = i
                break

            if next_i is None:
                raise ValueError('Trying to interpolate a point outside the series')

        else:
            if not self.series[prev_i].t < at < self.series[next_i].t:
                raise ValueError('Trying to interpolate a point outside the prev_i / next_i interval')

        # Shortcuts
        prev_point = self.series[prev_i]
        next_point = self.series[next_i]

        # Compute the increment with respect to the interpolation
        interpolated_data = self.series[0].data.__class__()

        for label in self.series.data_labels():
            interpolated_data[label] = prev_point.data[label] + ((next_point.data[label] - prev_point.data[label]) /2)

        return interpolated_data






# -*- coding: utf-8 -*-
"""Plotting engines."""

import os
import uuid
import datetime
from propertime.utils import dt_from_s, str_from_dt, dt_from_str, s_from_dt
from .units import TimeUnit
from .utils import is_numerical, os_shell, PFloat
try:
    from pyppeteer.chromium_downloader import download_chromium,chromium_executable
    image_plot_support=True
except ImportError:
    image_plot_support=False

# Setup logging
import logging
logger = logging.getLogger(__name__)

# Setup come configuration
AGGREGATE_THRESHOLD = int(os.environ.get('AGGREGATE_THRESHOLD', 10000))
DEFAULT_PLOT_TYPE = os.environ.get('DEFAULT_PLOT_TYPE', None)
if DEFAULT_PLOT_TYPE:
    if DEFAULT_PLOT_TYPE == 'interactive':
        logger.debug('Setting default plot type to "interactive"')
        DEFAULT_PLOT_AS_IMAGE = False
    elif DEFAULT_PLOT_TYPE == 'image':
        logger.debug('Setting default plot type to "image"')
        DEFAULT_PLOT_AS_IMAGE = True
    else:
        raise ValueError('Unknown plot type "{}" for DEFAULT_PLOT_TYPE'.forat(DEFAULT_PLOT_TYPE))
else:
    DEFAULT_PLOT_AS_IMAGE=False

#=================
#   Utilities
#=================

# Colors are based on the Tab 10 colormap (https://matplotlib.org/stable/users/explain/colors/colormaps.html#qualitative)

_data_index_colors = [
    (31, 119, 180),  # Blue
    (44, 160, 44),   # Green
    (214, 39, 40),   # Red
    (148, 103, 189), # Purple
    (140, 86, 75),   # Brown
    (127, 127, 127), # Gray
    (188, 189, 34),  # Olive
    (23, 190, 207)   # Cyan
]

_data_label_colors = [
    (0,128,128),      # Dygraphs green
    (148, 103, 189),  # Purple
    (44, 160, 44),    # Green
    (140, 86, 75),    # Brown
    (227, 119, 194),  # Pink
    (127, 127, 127),  # Gray
    (188, 189, 34),   # Olive
    (23, 190, 207),   # Cyan
    (214, 39, 40),    # Red
    (255, 127, 14),   # Orange
    (31, 119, 180),   # Blue
    # Extra colors (non-tab10) follows
    (77, 175, 74),    # Green
    (166, 86, 40),    # Brown
    (0,0,0),          # Black
    (255, 255, 51),   # Yellow
    (247, 129, 191),  # Pink
    (153, 153, 153),  # Gray
    (255, 127, 0),    # Orange
    (55, 126, 184),   # Blue
    (228, 26, 28),    # Red
]

def _utc_fake_s_from_dt(dt):
    dt_str = str_from_dt(dt)
    if '+' in dt_str:
        return s_from_dt(dt_from_str(dt_str.split('+')[0]+'+00:00'))
    elif '-' in dt_str:
        return s_from_dt(dt_from_str(dt_str.split('-')[0]+'+00:00'))
    else:
        raise ValueError('Cannot convert to fake UTC epoch a datetime without a timezone')

def _check_data_for_plot(data):
    if is_numerical(data):
        pass
    elif isinstance(data, list):
        for _,item in enumerate(data):
            if not is_numerical(item):
                raise Exception('Don\'t know how to plot item "{}" of type "{}"'.format(item, item.__class__.__name__))

    elif isinstance(data, dict):
        for _,item in data.items():
            if not is_numerical(item):
                raise Exception('Don\'t know how to plot item "{}" of type "{}"'.format(item, item.__class__.__name__))
    else:
        raise Exception('Don\'t know how to plot data "{}" of type "{}"'.format(data, data.__class__.__name__))

def _to_dg_time(dt):
    if dt.microsecond:
        return 'new Date(Date.UTC({}, {}, {}, {}, {}, {}, {}))'.format(dt.year, dt.month-1, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond/1000)
    else:
        return 'new Date(Date.UTC({}, {}, {}, {}, {}, {}))'.format(dt.year, dt.month-1, dt.day, dt.hour, dt.minute, dt.second)

def _to_dg_data(serie, data_labels_to_plot, data_indexes_to_plot, full_precision, aggregate_by=0, mark=None, pfloats=False, lower=0.01, upper=0.99):
    from .datastructures import DataTimePoint, DataTimeSlot

    dg_data=''
    first_t  = None
    last_t   = None
    labels = None
    global_min = None
    global_max = None

    for i, item in enumerate(serie):

        # Offset
        i=i+1

        # Set labels and support vars
        if not labels:
            labels = data_labels_to_plot
            if not labels:
                labels = [None]
            data_sums = [0 for label in labels]
            data_mins = [None for label in labels]
            data_maxs = [None for label in labels]
            index_sums = [0 for index in data_indexes_to_plot]


        #====================
        #  Aggregation
        #====================
        if aggregate_by:

            # Set first timestamp
            if first_t is None:
                first_t = item.t

            # Define data
            if labels:
                datas=[]
                for j, label in enumerate(labels):
                    if label is None:
                        datas.append(item.data)
                    else:
                        datas.append(item._data_by_label(label))
            else:
                datas = [item.data]

            # Loop over data labels, including the "None" label if we have no labels (float data), and add data
            for j, data in enumerate(datas):

                # Sum
                data_sums[j] += data

                # Global min
                if global_min is None:
                    global_min = data
                else:
                    if data < global_min:
                        global_min = data

                # Global max
                if global_max is None:
                    global_max = data
                else:
                    if data > global_max:
                        global_max = data

                # Min
                if data_mins[j] is None:
                    data_mins[j] = data
                else:
                    if data < data_mins[j]:
                        data_mins[j] = data

                # Max
                if data_maxs[j] is None:
                    data_maxs[j] = data
                else:
                    if data > data_maxs[j]:
                        data_maxs[j] = data

            # Loop over series data_indexes and add data
            for j, index in enumerate(data_indexes_to_plot):
                try:
                    index_value = item.data_indexes[index]
                    if index_value is not None:
                        index_sums[j] += index_value
                except KeyError:
                    pass

            # Dump aggregated data?
            if (i!=0)  and ((i % aggregate_by)==0):
                if issubclass(serie.item_type, DataTimePoint):
                    last_t = item.t
                elif issubclass(serie.item_type,DataTimeSlot):
                    last_t = item.end.t
                aggregation_t = first_t + ((last_t-first_t) /2)
                data_part=''

                # Data
                for i, label in enumerate(labels):
                    avg = data_sums[i]/aggregate_by
                    if full_precision:
                        data_part+='[{},{},{}],'.format(data_mins[i], avg, data_maxs[i])
                    else:
                        data_part+='[{:.2f},{:.2f},{:.2f}],'.format(data_mins[i], avg, data_maxs[i])

                # Data indexes
                for i, index in enumerate(data_indexes_to_plot):
                    if index_sums[i] is not None:
                        aggregated_index_value = index_sums[i]/aggregate_by
                        if full_precision:
                            data_part+='[0,{},{}],'.format(aggregated_index_value,aggregated_index_value)
                        else:
                            data_part+='[0,{:.4f},{:.4f}],'.format(aggregated_index_value,aggregated_index_value)
                    else:
                        data_part+='[null,null,null],'

                # Remove last comma
                data_part = data_part[0:-1]

                # Add to dg_data
                dg_data += '[{},{}],'.format(_to_dg_time(dt_from_s(aggregation_t, tz=item.tz)), data_part)

                # Reset averages
                data_sums = [0 for label in labels]
                data_mins = [None for label in labels]
                data_maxs = [None for label in labels]
                index_sums = [0 for index in data_indexes_to_plot]
                first_t = None

        #====================
        #  No aggregation
        #====================
        else:

            # Loop over data labels, including the "None" label if we have no labels (float data), and add data
            data_part=''
            for label in labels:
                if label is None:
                    data = item.data
                else:
                    data = item._data_by_label(label)

                if full_precision:
                    if pfloats:
                        if isinstance(data, PFloat):
                            data_part += '[{},{},{}],'.format(data.distf().dist_obj.ppf(lower, **data.dist['params']),
                                                              data,
                                                              data.distf().dist_obj.ppf(upper, **data.dist['params']))
                        else:
                            data_part += '[{},{},{}],'.format(data, data, data)
                    else:
                        data_part += '{},'.format(data)
                else:
                    if pfloats:
                        if isinstance(data, PFloat):
                            data_part += '[{:.2f},{:.2f},{:.2f}],'.format(data.distf().dist_obj.ppf(lower, **data.dist['params']),
                                                                          data,
                                                                          data.distf().dist_obj.ppf(upper, **data.dist['params']))
                        else:
                            data_part += '[{:.2f},{:.2f},{:.2f}],'.format(data,data,data)
                    else:
                        data_part += '{:.2f},'.format(data)

                # Global min
                if global_min is None:
                    global_min = data
                else:
                    if data < global_min:
                        global_min = data

                # Global max
                if global_max is None:
                    global_max = data
                else:
                    if data > global_max:
                        global_max = data

            # Loop over series data_indexes and add data
            for index in data_indexes_to_plot:
                try:
                    index_value = item.data_indexes[index]
                    if index_value is not None:
                        if full_precision:
                            if pfloats:
                                data_part+='[{},{},{}],'.format(index_value,index_value,index_value)
                            else:
                                data_part+='{},'.format(index_value)
                        else:
                            if pfloats:
                                data_part+='[{:.4f},{:.4f},{:.4f}],'.format(index_value,index_value,index_value)
                            else:
                                data_part+='{:.4f},'.format(index_value)
                    else:
                        data_part+='null,'
                except KeyError:
                    data_part+='null,'

            # Remove last comma
            data_part = data_part[0:-1]

            # Add to dg_data
            dg_data += '[{},{}],'.format(_to_dg_time(dt_from_s(item.t, tz=item.tz)), data_part)

    if dg_data.endswith(','):
        dg_data = dg_data[0:-1]

    return global_min, global_max, '['+dg_data+']'


#=================
#  Dygraphs plot
#=================

def dygraphs_plot(series, data_labels='all', data_indexes='all', aggregate=None, aggregate_by=None, full_precision=False,
                  color=None, data_label_colors='auto', data_index_colors='auto', height=None, image=DEFAULT_PLOT_AS_IMAGE,
                  image_resolution='auto', html=False, save_to=None, mini_plot='auto', value_range='auto', minimal_legend=False,
                  title=None, mark=None, mark_title=None, mark_color='auto', probabilistic='auto', probability_interval=[0.05, 0.95],
                  legacy='auto'):
    """Plot a time series using Dygraphs interactive plots.

       Args:
           series(TimeSeries): the time series to plot.
           data_labels(list): a list of data_labels to plot. By default set to all the data labels of the series.
           data_indexes(list): a list of data_indexes as the ``data_loss``, ``data_reconstructed`` etc.
                               to plot. By default set to all the data indexes of the series. To disable
                               plotting data indexes entirely, use None or an empty list.
           aggregate(bool): if to aggregate the time series, in order to speed up plotting.
                            By default, above 10000 data points the time series starts to
                            get aggregated by a factor of ten for each order of magnitude.
           aggregate_by(int): a custom aggregation factor.
           full_precision(bool): if to use (nearly) full precision using 6 significant figures instead of the
                                 automatic rounding. Defaulted to false.
           color(str): the (HTML) color of the time series in the plot (for univariate time series).
           data_label_colors(list,dict): the (HTML) colors of the time series in the plot. If provided as list, mapping follows
                                         the ordering of the data labels, while if provided as a dictionary then the mapping can
                                         be explicit (as {"data label": "color"}).
           data_index_colors(list,dict): the (HTML) colors of the time series data indexes in the plot. If provided as list,
                                         mapping follows the ordering of the data labels, while if provided as a dictionary then
                                         the mapping can be explicit (as {"data label": "color"}).
           height(int): force a plot height, in the time series data units.
           image(bool): if to generate an image rendering of the plot instead of the default interactive one.
                        To generate the image rendering, a headless Chromium web browser is downloaded on the
                        fly in order to render the plot as a PNG image.
           image_resolution(str): the image resolution, if generating an image rendering of the plot. Automatically
                                  set between ``1280x380`` and ``1024x400``, depending on the time series legend and title.
           html(bool): if to return the HTML code for the plot instead of generating an interactive or image one.
                       Useful for embedding it in a website or for generating multi-plot HTML pages.
           save_to(str): a file name (including its path) where to save the plot. If the plot is generated as
                         interactive, then it is saved as a self-consistent HTML page which can be opened in a
                         web browser. If the plot is generated as an image, then it is saved in PNG format.
           mini_plot(bool): if to include the range selector mini plot. Always automatically included unless
                            saving the plot in image format.
           value_range(list): a value range for the y axes to force the plot to stick with, in the form ``[min, max]``.
           minimal_legend(bool): if to strip down the information in the legend to the very minimum.
           title(str): a title for the plot.
           mark(str): a mark, to be used for highlighting a portion of the plot. Required to be formatted as a list or tuple
                      with two elements, the first from where the mark has to start and the second where it has to end.
           mark_title(str): a tile for the mark, to be displayed in the legend.
           mark_color(str): a color for the mark, defaults to light yellow.
           probabilistic(bool, str): if to enable the probabilistic support in the plots. Defaults to 'auto'.
           probability_interval(list): the probability interval lower and upper bounds. Defaults to [0.05,0.95].
           legacy(bool): if to enable legacy mode (required for Jupyter Notebook < 7, never required for Jupyter Lab).
    """
    # Credits: the interactive plot is based on the work here: https://www.stefaanlippens.net/jupyter-custom-d3-visualization.html.

    if probabilistic == 'auto':
        pfloats=False
        last_item_data = series[-1].data
        if isinstance(last_item_data, list):
            for item in last_item_data:
                if isinstance(item, PFloat):
                    pfloats = True
                    break
        if isinstance(last_item_data, dict):
            for item in last_item_data.values():
                if isinstance(item, PFloat):
                    pfloats = True
                    break
    else:
        if not probabilistic:
            pfloats = False
        else:
            pfloats = True

    if probability_interval:
        try:
            probability_interval[0]
            probability_interval[1]
        except TypeError:
            raise TypeError('Unsupported non-indexable type "{}" for the probability_interval argument'.format(probability_interval.__class__.__name__))
        except IndexError:
            raise IndexError('The probability_interval argument must have two values')

        if probability_interval[0] >= probability_interval[1]:
            raise ValueError('The probability_interval upper bound must be greater than the lower')

        if probability_interval[0] <= 0:
            raise ValueError('The lower bound of the probability interval cannot be equal to zero (or lower)')

        if probability_interval[1] >= 1:
            raise ValueError('The upper bound of the probability interval cannot be equal to one (or higher)')
    else:
        if pfloats:
            raise ValueError('Cannot generate a probabilistic plot without a probability interval')

    try:
        from IPython.display import display, Javascript, HTML, Image
    except ModuleNotFoundError:
        pass

    if html and image:
        raise ValueError('Setting both image and html to True is not supported.')

    if html and save_to:
        raise ValueError('Setting html=True is not compatible with setting a save_to value.')

    if len(series)==0:
        raise Exception('Cannot plot empty series')

    if save_to:
        if image:
            if not save_to.endswith('.png'):
                logger.warning('You are saving to "{}" in image (PNG) format, but the file name does not end with ".png"'.format(save_to))
        else:
            if not save_to.endswith('.html'):
                logger.warning('You are saving to "{}" in interactive (HTML) format, but the file name does not end with ".html"'.format(save_to))

    if aggregate_by is None and not pfloats:
        # Set default aggregate_by if not explicitly set to something
        if len(series)  > AGGREGATE_THRESHOLD:
            aggregate_by = 10**len(str(int(len(series)/float(AGGREGATE_THRESHOLD))))
        else:
            aggregate_by = None

    if aggregate is not None:
        # Handle the case where we are required not to aggregate
        if not aggregate:
            aggregate_by = None

    if aggregate_by and pfloats:
        raise ValueError('Cannot generate a probabilistic plot with an aggregation. Please disable it with "aggregate=False".')

    if mini_plot=='auto':
        if image:
            mini_plot=False
        else:
            mini_plot=True

    # Handle series data_lables
    if data_labels == 'all':
        # Plot all the series data_labels
        data_labels_to_plot = series.data_labels()
    elif data_labels is None:
        data_labels_to_plot = []
    else:
        # Check that the data_labels are of the right type and that are present in the series data_labels
        data_labels_to_plot = []
        if not isinstance(data_labels, list):
            raise TypeError('The "data_labels" argument must be a list or the magic word "all" (got type "{}")'.format(data_labels.__class__.__name__))
        for label in data_labels:
            if not isinstance(label, str):
                raise TypeError('The "data_labels" list items must be string (got type "{}")'.format(label.__class__.__name__))
            if not label in series.data_labels():
                raise ValueError('The data label "{}" is not present in the series data labeles (choices are: {})'.format(label, series.data_labels()))
            data_labels_to_plot.append(label)

    # Handle series data_indexes
    if data_indexes == 'all':
        # Plot all the series data_indexes
        data_indexes_to_plot = series._all_data_indexes()
    elif data_indexes is None:
        data_indexes_to_plot = []
    else:
        # Check that the data_indexes are of the right type and that are present in the series data_indexes
        data_indexes_to_plot = []
        if not isinstance(data_indexes, list):
            raise TypeError('The "data_indexes" argument must be a list or the magic word "all" (got type "{}")'.format(data_indexes.__class__.__name__))
        for index in data_indexes:
            if not isinstance(index, str):
                raise TypeError('The "data_indexes" list items must be string (got type "{}")'.format(index.__class__.__name__))
            if not index in series._all_data_indexes():
                raise ValueError('The data index "{}" is not present in the series data indexes (choices are: {})'.format(index, series._all_data_indexes()))
            data_indexes_to_plot.append(index)

    # Handle colors
    if data_label_colors == 'auto':
        data_label_colors = _data_label_colors
    if data_index_colors == 'auto':
        data_index_colors = _data_index_colors
        custom_data_index_colors = False
    else:
        custom_data_index_colors = True

    # Checks
    from .datastructures import DataTimePoint, DataTimeSlot
    if issubclass(series.item_type, DataTimePoint):
        stepPlot_value   = 'false'
        drawPoints_value = 'true'
        if aggregate_by:
            legend_pre = 'Point (aggregated by {}) at '.format(aggregate_by)
        else:
            legend_pre = 'Point at '
    elif issubclass(series.item_type, DataTimeSlot):
        if isinstance(series.resolution, TimeUnit):
            series_unit_string = str(series.resolution)
        else:
            series_unit_string = str(series.resolution) #.split('.')[0]+'s'
        stepPlot_value   = 'true'
        drawPoints_value = 'false'
        if aggregate_by:
            legend_pre = 'Slot of {}x{} (aggregated) starting at '.format(aggregate_by, series_unit_string)
        else:
            legend_pre = 'Slot of {} starting at '.format(series_unit_string)

    else:
        raise Exception('Don\'t know how to plot an object of type "{}"'.format(series.__class__.__name__))

    # Set legend points/slots time zone
    legend_tz = ' ({})'.format(series.tz)

    if minimal_legend:
        legend_pre = ''
        legend_tz = ''

    # Set series description
    if issubclass(series.item_type, DataTimePoint):
        if aggregate_by:
            series_desc = 'Time series of #{} points at {}, aggregated by {}'.format(len(series), series._resolution_string, aggregate_by)
        else:
            series_desc = 'Time series of #{} points at {}'.format(len(series), series._resolution_string)

    elif issubclass(series.item_type, DataTimeSlot):
        if aggregate_by:
            series_desc = 'Time series of #{} slots of {}, aggregated by {}'.format(len(series), series_unit_string, aggregate_by)
        else:
            series_desc = 'Time series of #{} slots of {}'.format(len(series), series_unit_string)

    else:
        series_desc = series.__class__.__name__

    if minimal_legend:
        series_desc = 'Legend'

    if aggregate_by:
        logger.info('Aggregating by "{}" for improved plotting'.format(aggregate_by))

    # Define div name
    graph_id = 'graph_' + str(uuid.uuid4()).replace('-', '')
    graph_div_id  = graph_id + '_plotarea'
    legend_div_id = graph_id + '_legend'

    # Do we have to show an highlight?
    mark_color_plot = 'rgba(255, 255, 102, .5)' if mark_color == 'auto' else mark_color
    mark_color_legend = 'rgba(255, 255, 102, .6)' if mark_color == 'auto' else mark_color
    if mark:
        #logger.info('Found series highlight and showing it')
        if not mark_title:
            mark_title = 'mark'
        series_mark_html = '<span style="background:{};">&nbsp;{}&nbsp;</span>'.format(mark_color_legend, mark_title)
    else:
        series_mark_html=''

    # Dygraphs Javascript
    dygraphs_javascript = """

// Clone the interaction model so that we do not interfere with the original dblclick method that might cause a recursive call
var touchFriendlyInteractionModel = Object.assign({}, Dygraph.defaultInteractionModel);

// Disable touch interaction (moves by default, we don't want that)
function nullfunction(){}
touchFriendlyInteractionModel.touchstart=nullfunction
touchFriendlyInteractionModel.touchend=nullfunction
touchFriendlyInteractionModel.touchmove=nullfunction

// Legend formatter taken and adapted from dygraph.js
function legendFormatter(data) {
      var g = data.dygraph;

      if (g.getOption('showLabelsOnHighlight') !== true) return '';

      var data_indexes = """+str(series._all_data_indexes())+""";
      var sepLines = g.getOption('labelsSeparateLines');
      var first_data_index = true;
      var html;

      if (typeof data.x === 'undefined') {
        if (g.getOption('legend') != 'always') {
          return '';
        }
        // We are here when there's no selection and {legend: 'always'} is set
        html = '"""+series_desc+""": ';
        for (var i = 0; i < data.series.length; i++) {
          var series = data.series[i];

          // Skip series not visible
          if (!series.isVisible) continue;

          // Also skip Timeseria-specific series
          if (series.label=='data_mark') continue;

          // Add some initial stuff
          if (html !== '') html += sepLines ? '<br/>' : ' ';

          if (data_indexes.includes(series.label)){
              html += "<span style='margin-left:15px; background: " + series.color + ";'>&nbsp;" + series.labelHTML + "&nbsp</span>, ";
          }
          else {
              html += "<span style='margin-left:15px; font-weight: bold; color: " + series.color + ";'>" + series.dashHTML + " " + series.labelHTML + "</span>, ";
          }

          // Remove last comma and space
          html = html.substring(0, html.length - 2);
        }
        html += ' &nbsp;&nbsp; """+series_mark_html +"""'
        return html;
      }

      // Find out if we have an active mark
      mark_active = false
      for (var i = 0; i < data.series.length; i++) {
        if ( data.series[i].label=='data_mark' && data.series[i].y==1 )  {
           mark_active = true;
        }
      }

      html = '""" + legend_pre + """' + data.xHTML + '""" + legend_tz + """:';
      for (var i = 0; i < data.series.length; i++) {

        var mark_active = false;

        if ( data.series[i].label=='data_mark' && data.series[i].y==1 )  {
           mark_active = true;
        }

        var series = data.series[i];

        // Skip series not visible
        if (!series.isVisible) continue;

        // Also skip Timeseria-specific series
        if (series.label=='data_mark') continue;

        if (sepLines) html += '<br>';
        var decoration = series.isHighlighted ? ' style="background-color: #000000"' : '';

        if (data_indexes.includes(series.label)){
            html += "<span" + decoration + "> <span style='background: " + series.color + ";'>&nbsp" + series.labelHTML + "&nbsp</span>:&#160;" + """+ ('series.yHTML*100' if full_precision else '(Math.round(series.yHTML *100 * 100) / 100)')  +""" + "%</span>, ";
        }
        else {
            html += "<span" + decoration + "> <b><span style='color: " + series.color + ";'>" + series.labelHTML + "</span></b>:&#160;" + series.yHTML + "</span>, ";
        }

      }
      // Remove last comma and space
      html = html.substring(0, html.length - 2);

      if (mark_active) {
          html += ' &nbsp;"""+ series_mark_html +"""'
      }

      return html;
    };
"""

    dygraphs_javascript += '\nnew Dygraph('
    dygraphs_javascript += 'document.getElementById("{}"),'.format(graph_div_id)

    # Check data for plot.
    _check_data_for_plot(series[0].data)

    # Loop over all the data labels
    labels=''
    if data_labels_to_plot:
        for data_label_to_plot in data_labels_to_plot:
            if isinstance(series[0].data, list) or isinstance(series[0].data, tuple):
                if len(data_labels_to_plot) == 1:
                    labels='value,'
                else:
                    labels += 'value {},'.format(data_label_to_plot)
            else:
                labels += '{},'.format(data_label_to_plot)
        # Remove last comma
        labels = labels[0:-1]
    else:
        labels='value'

    # Set index labels
    for index in data_indexes_to_plot:
        labels+=',{}'.format(index)

    # Prepare labels string list for Dygraphs
    labels_list = ['Timestamp'] + labels.split(',')

    # Get series data in Dygraphs format (and aggregate if too much data and find global min and max)
    global_min, global_max, dg_data = _to_dg_data(series,
                                                  data_labels_to_plot,
                                                  data_indexes_to_plot,
                                                  full_precision,
                                                  aggregate_by,
                                                  mark,
                                                  pfloats,
                                                  probability_interval[0],
                                                  probability_interval[1])
    if height:
        global_max = height
    if global_max != global_min:
        plot_min = str(global_min-((global_max-global_min)*0.1))
        plot_max = str(global_max+((global_max-global_min)*0.1))
    else:
        # Case where there is a straight line
        plot_min = str(global_min*0.99)
        plot_max = str(global_max*1.01)

    # Overwrite plot min and max?
    if value_range != 'auto':
        plot_min = str(value_range[0]*0.99)
        plot_max = str(value_range[1]*1.01)

    # Add data in dg (Dygraphs) format
    dygraphs_javascript += '{},'.format(dg_data)

    dygraphs_javascript += """{\
labels: """+str(labels_list)+""",
drawGrid: true,
drawPoints:"""+drawPoints_value+""",
strokeWidth: 1.5,
pointSize:1.5,
highlightCircleSize:4,
stepPlot: """+stepPlot_value+""",
fillGraph: false,
fillAlpha: 0.5,
colorValue: 0.5,
showRangeSelector: """+('true' if mini_plot else 'false')+""",
hideOverlayOnMouseOut: true,
interactionModel: touchFriendlyInteractionModel,
includeZero: false,
digitsAfterDecimal:2,
sigFigs: """ + ('6' if full_precision else 'null') + """,
showRoller: false,
legend: 'always',
labelsUTC: true,
legendFormatter: legendFormatter,
labelsDiv: '"""+legend_div_id+"""',
valueRange: ["""+plot_min+""", """+plot_max+"""],
zoomCallback: function() {
    this.updateOptions({zoomRange: ["""+plot_min+""", """+plot_max+"""]});
},
axes: {
  y2: {
    drawGrid: false,
    drawAxis: false,
    axisLabelWidth:0,
    independentTicks: true,
    valueRange: [0,1.01],
  }
},
animatedZooms: true,"""

    # Variable fill alpha in case of aggregated plot or not (for the area)
    # They have to be differentiated due different transparency policies in Dygraphs
    if aggregate_by:
        default_area_alpha_value  = 0.21
    else:
        default_area_alpha_value  = 0.4

    # Data indexes series start
    dygraphs_javascript += """
     series: {"""

    data_index_colors_count = 0
    for i, data_index_to_plot in enumerate(data_indexes_to_plot):
        if custom_data_index_colors:
            if isinstance(data_index_colors, list):
                data_index_color = data_index_colors[i%len(data_index_colors)]
            else:
                data_index_color = data_index_colors[data_index_to_plot]
            legend_alpha_value = 0.4
            area_alpha_value = default_area_alpha_value
        else:
            # Special data indexes
            if data_index_to_plot == 'data_reconstructed':
                data_index_color = (227, 168, 237) # Violet
                legend_alpha_value = 0.5
                area_alpha_value = 0.6 # Fixed in this case

            elif data_index_to_plot == 'data_loss':
                data_index_color = (255,128,128) # Red
                legend_alpha_value = 0.4
                area_alpha_value = default_area_alpha_value + 0.1 # Denser in this case

            elif data_index_to_plot == 'forecast':
                data_index_color = (255, 255, 102) # Yellow
                legend_alpha_value = 0.6
                area_alpha_value = default_area_alpha_value + 0.1 # Denser in this case

            elif data_index_to_plot in ['anomaly', 'anomaly_index']:
                data_index_color = (255, 140, 0) # Dark orange
                legend_alpha_value = 0.7
                area_alpha_value = 0.6 # Fixed in this case

            # Standard data indexes
            else:
                data_index_color = data_index_colors[data_index_colors_count%len(data_index_colors)]
                legend_alpha_value = 0.4
                area_alpha_value = default_area_alpha_value
                data_index_colors_count += 1

        # Assemble the color
        if isinstance(data_index_color, str):
            html_data_index_color = data_index_color
        else:
            html_data_index_color = 'rgba({},{},{},{})'.format(data_index_color[0],
                                                               data_index_color[1],
                                                               data_index_color[2],
                                                               legend_alpha_value)

        # Now create the entry for this data index
        dygraphs_javascript += """
        '"""+data_index_to_plot+"""': {
         axis: 'y2',
         drawPoints: false,
         strokeWidth: 0,
         highlightCircleSize:0,
         fillGraph: true,
         fillAlpha: """+str(area_alpha_value)+""", // Alpha here is used for the area
         color: '"""+html_data_index_color+"""' // Alpha here is used for the legend
       },"""

    # Add all non-index series to be included in the miniplot
    for label in data_labels_to_plot:
        dygraphs_javascript += """
           '"""+str(label)+"""': {
             showInRangeSelector:true
           },"""

    # Data indexes series end
    dygraphs_javascript += """
     },"""

    # Set data label colors
    colors = "colors: ["
    for i, data_label_to_plot in enumerate(data_labels_to_plot):
        if i==0 and color:
            colors += "'{}',".format(color)
        else:
            if isinstance(data_label_colors, list):
                data_label_color = data_label_colors[i%len(data_label_colors)]
            else:
                data_label_color = data_label_colors[data_label_to_plot]
            if isinstance(data_label_color, str):
                colors += "'{}',".format(data_label_color)
            else:
                colors += "'rgb({},{},{})',".format(data_label_color[0],
                                                    data_label_color[1],
                                                    data_label_color[2])
    colors = colors[:-1] # Remove trailing comma
    colors += "],"
    dygraphs_javascript += colors

    # Handle series mark if set
    if mark:

        # Check that we know how to use this mark
        if not isinstance(mark[0], datetime.datetime):
            raise TypeError('Mark start/end must be datetime objects')
        if not isinstance(mark[1], datetime.datetime):
            raise TypeError('Mark start/end marks be datetime objects')

        # Convert the mark to fake epoch milliseconds
        mark_start = _utc_fake_s_from_dt(mark[0])*1000
        mark_end   = _utc_fake_s_from_dt(mark[1])*1000

        # Add js code
        dygraphs_javascript += 'underlayCallback: function(canvas, area, g) {'
        dygraphs_javascript += 'var bottom_left = g.toDomCoords({}, -20);'.format(mark_start)
        dygraphs_javascript += 'var top_right = g.toDomCoords({}, +20);'.format(mark_end)
        dygraphs_javascript += 'var left = bottom_left[0];'
        dygraphs_javascript += 'var right = top_right[0];'
        dygraphs_javascript += 'canvas.fillStyle = "{}";'.format(mark_color_plot)
        dygraphs_javascript += 'canvas.fillRect(left, area.y, right - left, area.h);'
        dygraphs_javascript += '}'

    if aggregate_by or pfloats:
        # Add a comma, just if the previous section kicked-in
        if dygraphs_javascript[-1] != ',':
            dygraphs_javascript += ','
        dygraphs_javascript += 'customBars: true'
    dygraphs_javascript += '})'

    rendering_javascript = """
require.undef('"""+graph_id+"""'); // Helps to reload in Jupyter
define('"""+graph_id+"""', ['dgenv'], function (Dygraph) {

    function draw(div_id) {
    """ + dygraphs_javascript +"""
    }
    return draw;
});
"""

    STATIC_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/static/'


    if image or save_to or html:

        # By defualt we show the plot
        show_plot = True

        # Generate random UUID
        rnd_uuid=str(uuid.uuid4())

        # If a destination was set, we do not show the plot.
        if save_to:
            show_plot = False

        # Set destination file values
        if save_to:
            if image:
                # Dump as image
                html_dest = '/tmp/{}.html'.format(rnd_uuid)
                png_dest = save_to
            else:
                # Interactive, dump as html
                html_dest = save_to
                png_dest=None
        else:
            if image:
                # Dump as image
                html_dest = '/tmp/{}.html'.format(rnd_uuid)
                png_dest = '/tmp/{}.png'.format(rnd_uuid)
            else:
                # Will never get here, directly rendered in iPython
                pass

        # Start building HTML content
        return_html_code = html
        html_content = ''
        if not return_html_code:
            html_content += '<html>\n<head>\n<meta charset="UTF-8">'
            with open(STATIC_DATA_PATH+'/js/dygraph-2.1.0.min.js') as dg_js_file:
                html_content += '\n<script type="text/javascript">'+dg_js_file.read()+'</script>\n'
            with open(STATIC_DATA_PATH+'css/dygraph-2.1.0.css') as dg_css_file:
                html_content += '\n<style>'+dg_css_file.read()+'</style>\n'
            html_content += '</head>\n<body style="font-family:\'Helvetica Neue\', Helvetica, Arial, sans-serif; font-size:1.0em">\n'
            if title:
                html_content += '<div style="text-align: center; margin-top:15px; margin-bottom:15px; font-size:1.2em"><h3>{}</h3></div>'.format(title)
            html_content += '<div style="height:36px; padding:0; margin-left:0px; margin-top:10px">\n'
        else:
            if title:
                html_content += '<div style="text-align: center; margin-top:15px; margin-bottom:15px; font-size:1.2em"><h3>{}</h3></div>'.format(title)
        html_content += '<div id="{}" style="width:100%"></div>\n'.format(legend_div_id)
        html_content += '<div id="{}" style="width:100%; margin-right:0px"></div>\n'.format(graph_div_id)
        if not return_html_code:
            html_content += '</body>\n</html>\n'
        html_content += '<script>{}\n</script>\n'.format(dygraphs_javascript)

        # Dump html to file
        try:
            with open(html_dest, 'w') as f:
                f.write(html_content)
        except UnboundLocalError:
            pass

        # Also render if not interactive mode
        if image:

            if not image_plot_support:
                raise ValueError('Sorry, image plots are not supported since the package "pyppeteer" could be imported')

            _chrom_executable = None
            _headless_mode_switch = ''

            # Is there a system Chrome or Chromium we can use?
            potential_chrom_executables = ['chromium', 'chromium-browser', 'google-chrome',
                                           '/Applications/Google\\ Chrome.app/Contents/MacOS/Google\\ Chrome']

            for potential_chrom_executable in potential_chrom_executables:
                out = os_shell('{} --version'.format(potential_chrom_executable), capture=True)
                if out.exit_code != 0:
                    continue
                else:
                    version = out.stdout.replace('Google', '').replace('Chrome', '').replace('Chromium', '').strip().split(' ')[0]
                    version_major = int(version.split('.')[0])
                    if version_major >= 59:
                        logger.debug('Found usable Chrom* executable: "{}"'.format(potential_chrom_executable))
                        _chrom_executable = potential_chrom_executable
                        if version_major >= 112:
                            _headless_mode_switch='=new'
                        break
                    else:
                        continue
                    logger.debug('Found {}'.format(potential_chrom_executable))

            # Otherwise, rely on Pyppeteer's downloading it
            if not _chrom_executable:
                logger.debug('Checking Pyppeteer\'s Chromium presence for rendering image-based plots...')
                if not chromium_executable().exists():
                    logger.info('Downloading Pyppeteer\'s Chromium for rendering image-based plots...')
                    # Disable logging for Pyppeteer
                    pyppeteer_logger = logging.getLogger('pyppeteer')
                    pyppeteer_logger.setLevel(logging.CRITICAL)
                    download_chromium()
                    logger.info('Done. Will not be required anymore on this system.')
                else:
                    logger.debug('Ok, Chromium present.')

                # Convert executable path to string and handle white spaces if present
                _chrom_executable = str(chromium_executable()).replace(' ', '\\ ')

            # Check path exists and is writable or Chrom* will risk to hang up
            png_dest_path = '/'.join(png_dest.split('/')[0:-1])
            if png_dest_path.strip():
                if not os.path.exists(png_dest_path):
                    try:
                        os.makedirs(png_dest_path)
                    except:
                        raise IOError('Path "{}" does not exists and cannot create it'.format(png_dest_path))
                try:
                    with open(png_dest, 'a'):
                        os.utime(png_dest, None)
                except:
                    raise IOError('Cannot write to "{}"'.format(png_dest))
                finally:
                    try:
                        os.remove(png_dest)
                    except:
                        pass

            # Render HTML to image
            if image_resolution == 'auto':
                if title:
                    image_resolution = '1280x490' if _headless_mode_switch else '1280x430'
                else:
                    image_resolution = '1280x440' if _headless_mode_switch else '1280x380'
            resolution = image_resolution.replace('x', ',')
            command = '{} --no-sandbox --headless{} --disable-gpu --window-size={} --screenshot={} {}'.format(_chrom_executable, _headless_mode_switch, resolution, png_dest, html_dest)
            logger.debug('Executing "{}"'.format(command))
            out = os_shell(command, capture=True)

            if not out.exit_code == 0:
                raise OSError('Error: {}'.format(out.stderr))

            if show_plot:
                return (Image(filename=png_dest))

        if html:
            return html_content

        # Log
        if save_to:
            if image:
                logger.info('Saved image plot in PNG format to "{}"'.format(png_dest))
            else:
                logger.info('Saved interactive plot in HTML format to "{}"'.format(html_dest))

    else:

        # Version/mode mapping:
        # Jupyter Notebook 5: legacy mode
        # Jupyter Notebook 6: legacy approach
        # Jupyter Notebook 7: normal mode
        # Jupyetr lab 3: normal mode
        # Jupyter Lab 4: normal mode

        if legacy=='auto':

            legacy = False

            from .utils import _detect_notebook_major_version
            if _detect_notebook_major_version() < 7:
                logger.warning('A Jupyter Notebook < 7 installation has been detect on this system. ' +
                               'If the plot does not show, enable the legacy plotting mode with legacy=True and restart the Kernel. ' +
                               'This is required for Jupyter Notebook < 7, but not for older versions of Jupyer Lab. However, it is not possible to detect ' +
                               'if running on Jupyter Notebook or Jupyter Lab, hence this warning. To suppress it, explicitly set legacy to True or False.')

        if legacy:
            # Load Dygraphs Javascript library
            # https://raw.githubusercontent.com/sarusso/Timeseria/develop/timeseria/static/js/dygraph-2.1.0.min
            display(Javascript("require.config({paths: {dgenv: 'https://cdnjs.cloudflare.com/ajax/libs/dygraph/2.1.0/dygraph.min'}});"))

            # TODO: load the Dygraphs Javascript locally, maybe with something like:
            #with open(STATIC_DATA_PATH+'js/dygraph-2.1.0.min.js') as dy_js_file:
            #    display(Javascript(dy_js_file.read()))

            # Load Dygraphs CSS
            with open(STATIC_DATA_PATH+'css/dygraph-2.1.0.css') as dg_css_file:
                display(HTML('<style>'+dg_css_file.read()+'</style>'))

            # Load Dygraphs Javascript and html code
            display_html = ''
            if title:
                display_html += '<div style="text-align: center; margin-bottom:20px; font-size:0.95em"><h3>{}</h3></div>'.format(title)
            display_html += '<div style="height:36px; padding:0; margin-left:0px; margin-top:10px">'
            display_html += '''<div id="'''+legend_div_id+'''" style="width:100%"></div></div>
                            <div id="'''+graph_div_id+'''" style="width:100%; margin-right:0px"></div>'''

            display(Javascript(rendering_javascript))
            display(HTML(display_html))

            # Render the graph (this is the "entrypoint")
            display(Javascript("""
                (function(element){
                    require(['"""+graph_id+"""'], function("""+graph_id+""") {
                        """+graph_id+"""(element.get(0), '%s');
                    });
                })(element);
            """ % (graph_div_id)))

        else:
            if title:
                display_html = '<div style="height:370px">'
            else:
                display_html = '<div style="height:360px">'
            with open(STATIC_DATA_PATH+'/js/dygraph-2.1.0.min.js') as dg_js_file:
                display_html += '\n<script type="text/javascript">'+dg_js_file.read()+'</script>\n'
            with open(STATIC_DATA_PATH+'css/dygraph-2.1.0.css') as dg_css_file:
                display_html += '\n<style>'+dg_css_file.read()+'</style>\n'
            if title:
                display_html += '<div style="text-align: center; margin-top:15px; margin-bottom:15px; font-size:1.0em"><h3>{}</h3></div>'.format(title)
            display_html += '<div style="height:36px; padding:0; margin-left:0px; margin-top:10px">\n'
            display_html += '<div id="{}" style="width:100%"></div>\n'.format(legend_div_id)
            display_html += '<div id="{}" style="width:100%; margin-right:0px"></div>\n'.format(graph_div_id)
            display_html += '<script>{}\n</script>\n'.format(dygraphs_javascript)
            display_html += '</div>'
            display(HTML(display_html))



#=================
# Matplotlib plot
#=================

def matplotlib_plot(series):
    """Plot a series using Matplotlib.

        Args:
            series(Series): the series to plot.
    """
    from matplotlib import pyplot
    pyplot.plot(series.to_df())
    pyplot.show()





#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='timeseria',
      version='2.1.0',
      description='An object-oriented time series processing library',
      long_description="""Timeseria is an object-oriented time series processing library which aims at making it easy to manipulate time series data and to build statistical 
and machine learning models on top of it.\n\nCheck out the GitHub project for more info: [Timeseria on GitHub](https://github.com/sarusso/Timeseria).\n\n
![Time series plot](https://raw.githubusercontent.com/sarusso/Timeseria/master/docs/altogether.png "Timeseria at work")""",
      long_description_content_type='text/markdown',
      url="https://github.com/sarusso/timeseria",
      author='Stefano Alberto Russo',
      author_email='stefano.russo@gmail.com',
      packages=['timeseria','timeseria.tests', 'timeseria.models'],
      package_data={
          'timeseria': ['static/css/*.css', 'static/js/*.js'],
          'timeseria.tests': ['test_data/csv/*.csv']
       },
      install_requires = [
                          'matplotlib >=2.1.2, <4.0.0',
                          'numpy >=1.19.5, <2.0.0',
                          'scikit-learn >=0.2.2, <2.0.0',
                          'pandas >=0.23.4, <2.0.0',
                          'chardet >=3.0.4, <4.0.0',
                          'convertdate >=2.1.2, <3.0.0',
                          'lunarcalendar >=0.0.9, <1.0.0',
                          'cython >=0.29.17, <1.0.0',
                          'requests >=2.20.0, <3.0.0',
                          'scipy >=1.5.4, <2.0.0',
                          'pyppeteer>=0.2.6, <1.0.0',
                          'fitter==1.7.0',
                          'propertime>=1.0.1, <2.0.0'
                          ],
      extras_require = {
                        'tensorflow': ['tensorflow >=2.0.0, <3.0.0'],
                        'tensorflow-gpu': ['tensorflow-gpu >=2.0.0, <3.0.0'],
                        'tensorflow-macos': ['tensorflow-macos >=2.0.0, <3.0.0'],
                        'tensorflow-aarch64': ['tensorflow-aarch64 >=2.0.0, <3.0.0'],
                        'prophet':['prophet >=1.1.1, <2.0.0'],
                        'arima': ['pmdarima >=1.8, <3.0.0', 'statsmodels >=0.14.0, <1.0.0']
                       },
      license='Apache License 2.0',
      license_files = ('LICENSE',),
    )
